#!/usr/bin/env python

# Kmax
# Copyright (C) 2012-2015 Paul Gazzillo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Find all compilation units given a list of directories.

import sys
import os
import glob
import re
import fnmatch
import argparse
import subprocess
try:
    import cPickle as pickle
except ImportError:  #Python3
    import pickle
import time

import kmaxtools.vcommon as CM

scriptpath = os.path.dirname(os.path.realpath(__file__))

starting_time = time.time()

def chgext(filename, f, t):
  if filename.endswith(f):
    return filename[:-2] + t

def otoc(filename):
  return chgext(filename, ".o", ".c")

def hostprog_otoc(filename):
  """host programs don't have a .o extension, but their components
  might if it's a composite"""
  if not filename.endswith(".o"):
    filename = filename + ".o"
  return chgext(filename, ".o", ".c")

def otoS(filename):
  return chgext(filename, ".o", ".S")

def ctoo(filename):
  return chgext(filename, ".c", ".o")

def print_set(s, name):
  for elem in s:
    print name,elem
  print "size",name,len(s)

argparser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description="""\
Find a set of configurations that covers all configuration variables in the
given Kbuild Makefile.
    """,
    epilog="""\
    """
    )
argparser.add_argument('makefile',
                       nargs="*",
                       type=str,
                       help="""the name of a Linux Makefile or subdir""")
argparser.add_argument('-D',
                       '--define',
                       action='append',
                       help="""\
define a make variable""")
argparser.add_argument('-x',
                       '--excludes-file',
                       help="""\
provides the excludes filename for reading and writing subdirectories that run \
without error""")
argparser.add_argument('-c',
                       '--correlate-c-files',
                       action="store_true",
                       help="""\
find corresponding .c files for the compilation units""")
argparser.add_argument('-C',
                       '--config-vars',
                       type=str,
                       help="""the name of a KConfigData file containing \
configuration variable data""")
argparser.add_argument('--no-aggregation',
                       action="store_true",
                       help="""\
only perform Kbuild evaluation, no aggregation and analysis""")
# argparser.add_argument('-g',
#                        '--get-presence-conditions',
#                        action="store_true",
#                        help="""\
# get presence conditions for each compilation units""")
argparser.add_argument('-B',
                       '--boolean-configs',
                       action="store_true",
                       help="""\
Treat all configuration variables as Boolean variables""")
# argparser.add_argument('-n',
#                        '--naive-append',
#                        action="store_true",
#                        help="""\
# use naive append behavior, which has more exponential explosion""")


args = argparser.parse_args()

if len(args.makefile) == 0:
  argparser.print_help()
  sys.exit(1)

toplevel_dirs = args.makefile

excludes = set()
if args.excludes_file != None:
  if os.path.exists(args.excludes_file):
    with open(args.excludes_file, "r") as f:
      excludes = pickle.load(f)

def covering_set(kbuild_dir,        # src directory to process
                 compilation_units, # updated with units
                 library_units,     # updated with lib units
                 hostprog_units,    # updated with hostprog units
                 unconfigurable_units,    # updated with
                                          # unconfigurable units
                 extra_targets,     # updated with extra targets
                 clean_files,       # updated with clean-files units
                 c_file_targets,    # updated with c-files from targets var
                 composites,        # updated with composites
                 broken,            # updated with kbuild files that
                                    # break kmax
                 presence_conditions):
  """Call the covering set program to find the list of compilation
  units and subdirectories added by the makefile in kbuild_dir"""
  global excludes

  if kbuild_dir in excludes:
    print "skipping", kbuild_dir
    return set()

  covering_set_args = [ "python", os.path.join(scriptpath, "kmax"),
                        "--case-study", "linux",
                        "--log_level", "0",
                        # "-p",  #tvn why use this flag ? 
                        # "-Dsrc=" + kbuild_dir,      # drivers/staging/wlags49_h25/, drivers/gpu/drm/nouveau/
                        # # TODO default to empty variable
                        # # "-Dobj=" + kbuild_dir,    # drivers/scsi/aic7xxx/
                        # # "-DKERNELDIR=",           # drivers/staging/wlags49_h25/
                        # # "-Dlibservices=",         # drivers/staging/tidspbridge/
                        # # "-DPWD=",                 # drivers/staging/rts5139/
                        # # "-DGCOV=",                # drivers/scsi/lpfc/
                        # # "-DDBGDEF=",              # drivers/net/wan/lmc/
                        # # "-DACPI_FUTURE_USAGE=",   # drivers/acpi/acpica/
                        ]

  # if args.define != None:
  #   for define in args.define:
  #     covering_set_args.append("-D" + define)

  # if args.config_vars:
  #   covering_set_args.append("-C" + args.config_vars)

  # if args.get_presence_conditions:
  #   covering_set_args.append("-g")

  if args.boolean_configs:
    covering_set_args.append("-B")

  # if args.naive_append:
  #   covering_set_args.append("-n")
  covering_set_args.append(kbuild_dir)


  sys.stderr.write("running covering_set: {}\n".format(' '.join(covering_set_args)))
  #CM.pause()
  print covering_set_args
  p = subprocess.Popen(covering_set_args,
                       stdout=subprocess.PIPE,
                       # stderr=subprocess.PIPE
                       )
  out, err = p.communicate()

  if p.returncode != 0:
    broken.add(kbuild_dir)
    return set()

  excludes.add(kbuild_dir)

  (new_compilation_units,
   new_library_units,
   new_hostprog_units,
   new_unconfigurable_units,
   new_extra_targets,
   new_clean_files,
   new_c_file_targets,
   new_pending_subdirectories,
   new_composites,
   new_presence_conditions) = pickle.loads(out)

  compilation_units.update(new_compilation_units)
  library_units.update(new_library_units)
  hostprog_units.update(new_hostprog_units)
  unconfigurable_units.update(new_unconfigurable_units)
  extra_targets.update(new_extra_targets)
  clean_files.update(new_clean_files)
  c_file_targets.update(new_c_file_targets)
  composites.update(new_composites)
  presence_conditions.update(new_presence_conditions)

  # if args.get_presence_conditions:
  #   # print presence condition information
  #   for unit_name, pc in new_unit_pcs:
  #     print "unit_pc", unit_name, pc
  #   for subdir_name, pc in new_subdir_pcs:
  #     print "subdir_pc", subdir_name, pc

  return new_pending_subdirectories

compilation_units = set()
subdirectories = set()
library_units = set()
hostprog_units = set()
unconfigurable_units = set()
extra_targets = set()
clean_files = set()
c_file_targets = set()
composites = set()
pending_subdirectories = set()
broken = set()
presence_conditions = set()

# find all compilation_units.  run covering_set.py until no more
# Kbuild subdirectories are left.
sys.stderr.write("running covering_set\n")
pending_subdirectories.update(args.makefile)
while len(pending_subdirectories) > 0:
  subdirectories.update(pending_subdirectories)
  pending_subdirectories.update(covering_set(pending_subdirectories.pop(),
                                             compilation_units,
                                             library_units,
                                             hostprog_units,
                                             unconfigurable_units,
                                             extra_targets,
                                             clean_files,
                                             c_file_targets,
                                             composites,
                                             broken,
                                             presence_conditions))

# if args.get_presence_conditions:
#   # already printed presence conditions.  don't do anything, so exit
#   exit(1)

if args.no_aggregation:
  print_set(toplevel_dirs, "toplevel_dirs")  # list of directories started from
  print_set(subdirectories, "subdirectory")  # subdirectory visited by kbuild
  print_set(composites, "composites")  # compilation units that are composites
  print_set(library_units, "library_units")  # library units referenced by kbuild
  print_set(hostprog_units, "hostprog_units")
  print_set(unconfigurable_units, "unconfigurable_units")
  print_set(extra_targets, "extra_targets")
  print_set(clean_files, "clean_files")
  print_set(c_file_targets, "c_file_targets")
  print_set(broken, "broken")
  print_set(presence_conditions, "presence_conditions")
  print "running_time", time.time() - starting_time
  exit(0)

sys.stderr.write("aggregating and analyzing covering_set data")

# find all subdirectories with source in them
used_subdirectory = set()
for unit in compilation_units:
  used_subdirectory.add(os.path.dirname(unit))

# find all .c files
all_c_files = set([])
for subdir in (subdirectories | used_subdirectory):
  # all_c_files.update([os.path.normpath(x) for x in glob.glob(os.path.join(subdir, "*.c"))])
  all_c_files.update([os.path.abspath(os.path.normpath(x)) for x in glob.glob(os.path.join(subdir, "*.c"))])

# find all compilation units without a corresponding .c file
unmatched_units = set()
asm_compilation_units = set()
for unit in compilation_units:
  c_file = otoc(unit)
  S_file = otoS(unit)
  if not os.path.isfile(c_file) and not os.path.isfile(S_file):
    unmatched_units.add(c_file)
  if os.path.isfile(S_file):
    asm_compilation_units.add(S_file)

# find all asm-offsets.c files, for these are compiled by the root
# Kbuild file into offsets
re_offsets_file = re.compile(r'arch/[^/]+/kernel/asm-offsets\.c')

offsets_files = [ x for x in all_c_files if re_offsets_file.match(x) ]

# find all .c files without a corresponding compilation unit, starting
# with all c files
unidentified_c_files = set(all_c_files)

# remove kernel compilation units
unidentified_c_files.difference_update([otoc(filename)
                                     for filename in compilation_units])

# remove library compilation units
unidentified_c_files.difference_update([otoc(filename)
                                     for filename in library_units])

# remove hostprog compilation units
unidentified_c_files.difference_update([hostprog_otoc(filename)
                                     for filename in hostprog_units])

# remove unconfigurable compilation units
unidentified_c_files.difference_update([hostprog_otoc(filename)
                                     for filename in unconfigurable_units])

# remove unconfigurable compilation units
unidentified_c_files.difference_update([hostprog_otoc(filename)
                                     for filename in extra_targets])

# remove unconfigurable compilation units
unidentified_c_files.difference_update([filename
                                     for filename in clean_files])

# remove asm-offsets.c files
unidentified_c_files.difference_update([filename
                                        for filename in offsets_files])

# get source files that include c files
included_c_files = set()
p = subprocess.Popen(r'find . -name "*.[c|h]" | xargs grep -H "^#.*include.*\.c[\">]"',
                     shell=True,
                     stdout=subprocess.PIPE)
out, _ = p.communicate()
for line in out.split("\n"):
  parts = line.split(":", 1)
  if parts is not None:
    infile = parts[0]
    if len(parts) > 1:
      m = re.search(r"\".*\.c\"", parts[1])
      if m is not None:
        included = m.group(0)[1:-1]
        included = os.path.join(os.path.dirname(infile), included)
        included = os.path.relpath(os.path.realpath(included))
        included = os.path.abspath(included)  # added to do abs paths
        included_c_files.add(included)

# only need the files in the current source subtree
included_c_files.intersection_update(all_c_files)

# remove .c files that aren't compilation units, because they are
# included in other c files
unidentified_c_files.difference_update(included_c_files)

# separate out .c files from the staging directory, which can be a
# mess
unidentified_staging_c_files = [ x for x in unidentified_c_files
                              if "drivers/staging" in os.path.dirname(x) ]

unidentified_c_files.difference_update(unidentified_staging_c_files)

# separate out .c files with "skeleton" in their name.  this is a
# heuristic to find example drivers that aren't actually compiled.
unidentified_skeleton_c_files = [ x for x in unidentified_c_files
                               if "skeleton" in os.path.basename(x) ]

unidentified_c_files.difference_update(unidentified_skeleton_c_files)

# separate out generators heuristically.  look for "mk" or
# "gen[^a-zA-Z]" in their name

# perhaps we can find generators a better way, e.g., by looking for
# the c file in the makefile

# look for unexpanded variables or function calls
re_unexpanded = re.compile(r'.*\$\(.*\).*')
unexpanded_compilation_units = [ x for x in compilation_units
                                 if re_unexpanded.match(x) ]
unexpanded_library_units = [ x for x in library_units
                             if re_unexpanded.match(x) ]
unexpanded_hostprog_units = [ x for x in hostprog_units
                              if re_unexpanded.match(x) ]
unexpanded_unconfigurable_units = [ x for x in unconfigurable_units
                                    if re_unexpanded.match(x) ]
unexpanded_extra_targets = [ x for x in extra_targets
                                 if re_unexpanded.match(x) ]
unexpanded_clean_files = [ x for x in clean_files
                             if re_unexpanded.match(x) ]
unexpanded_subdirectories = [ x for x in subdirectories
                              if re_unexpanded.match(x) ]

# remove compilation units with unexpanded variable names
unmatched_units.difference_update([ otoc(x)
                                    for x in unexpanded_compilation_units ])

# remove library units with unexpanded variable names
unmatched_units.difference_update([ otoc(x)
                                    for x in unexpanded_library_units ])

# remove hostprog units with unexpanded variable names
unmatched_units.difference_update([ hostprog_otoc(x)
                                    for x in unexpanded_hostprog_units ])

# remove unconfigurable units with unexpanded variable names
unmatched_units.difference_update([ hostprog_otoc(x)
                                    for x in unexpanded_unconfigurable_units ])

# remove extra targets with unexpanded variable names
unmatched_units.difference_update([ hostprog_otoc(x)
                                    for x in unexpanded_extra_targets ])

# remove clean targets with unexpanded variable names
unmatched_units.difference_update([ hostprog_otoc(x)
                                    for x in unexpanded_clean_files ])

# remove c files specified in the clean-files and in targets, since
# these can be auto-generated c files
generated_c_files = set([])
for c in (clean_files | c_file_targets):
  pattern = re.compile(fnmatch.translate(c))
  for filename in unmatched_units:
    if pattern.match(filename):
      generated_c_files.add(filename)
unmatched_units.difference_update(generated_c_files)

# cache the list of working kbuild files
if args.excludes_file != None:
  with open(args.excludes_file, "w") as f:
    pickle.dump(excludes, f, 0)

print_set(toplevel_dirs, "toplevel_dirs")  # list of directories started from
print_set(all_c_files, "all_c_files")  # all .c files in used and visited subdirectories
print_set(asm_compilation_units, "asm_compilation_units")  # compilation units with a .S file
print_set(subdirectories, "subdirectory")  # subdirectory visited by kbuild
print_set(used_subdirectory, "used_subdirectory")  # subdirectories containing compilation units
print_set(compilation_units, "compilation_units")  # compilation units referenced by kbuild
print_set(["{} {}".format(x, y.replace('\n', '').replace(' ', '')) for x, y in presence_conditions], "presence_conditions") # presence conditions for the compilation units and subdirectories
print_set(composites, "composites")  # compilation units that are composites
print_set(library_units, "library_units")  # library units referenced by kbuild
print_set(hostprog_units, "hostprog_units")
print_set(unconfigurable_units, "unconfigurable_units")
print_set(extra_targets, "extra_targets")
print_set(clean_files, "clean_files")
print_set(c_file_targets, "c_file_targets")
print_set(generated_c_files, "generated_c_files")
print_set(unmatched_units, "unmatched_units")
print_set(included_c_files, "included_c_files")
print_set(offsets_files, "offsets_files")
print_set(unidentified_c_files, "unidentified_c_files")
print_set(unidentified_staging_c_files, "unidentified_staging_c_files")
print_set(unidentified_skeleton_c_files, "unidentified_skeleton_c_files")
print_set(unexpanded_compilation_units, "unexpanded_compilation_units")
print_set(unexpanded_library_units, "unexpanded_library_units")
print_set(unexpanded_hostprog_units, "unexpanded_hostprog_units")
print_set(unexpanded_unconfigurable_units, "unexpanded_unconfigurable_units")
print_set(unexpanded_extra_targets, "unexpanded_extra_targets")
print_set(unexpanded_subdirectories, "unexpanded_subdirectories")
print_set(broken, "broken")
print "running_time", time.time() - starting_time



import re
import os
import pdb
trace = pdb.set_trace
import z3
try:
    import cPickle as pickle
except ImportError:  #Python3
    import pickle
import sys

import kmaxtools.vcommon as CM

import kmaxtools.settings
mlog = CM.getLogger(__name__, kmaxtools.settings.logger_level)

class FileAnalysis:
    @classmethod
    def chgext(cls, filename, f, t):
        if f is None or filename.endswith(f):
            return filename[:-2] + t
        
    @classmethod
    def hostprog_otoc(cls, filename):
        """host programs don't have a .o extension, but their components
        might if it's a composite"""
        if not filename.endswith('.o'):
            filename = filename + '.o'
        return cls.chgext(filename, '.o', '.c')

    @classmethod
    def otoc(cls, filename):
        return cls.chgext(filename, '.o', '.c')

    @classmethod
    def otoS(cls, filename):
        return cls.chgext(filename, '.o', '.S')

    @classmethod
    def ctoo(cls, filename):
        return chgext(filename, '.c', '.o')
    
    @classmethod
    def mkc(cls, filename):
        return cls.chgext(filename, None, '.c')

    @staticmethod
    def get_all_c_files(path):
        assert os.path.isdir(path)
        cmd = 'find {} -type f | grep "\.c$" | sort | uniq'.format(path)
        files, _ = CM.vcmd(cmd)
        files = set([os.path.normpath(x.strip('\n'))
                     for x in files.split()])
        return files
        
    @staticmethod
    def get_included_c_files(path):
        """get source files that include c files"""
        
        path = os.path.abspath(path)
        cmd = r'find {} -name "*.[c|h]" | xargs grep -H "^#.*include.*\.c[\">]"'.format(path)
        rs, _  = CM.vcmd(cmd)
        rs = [s.split(':', 1) for s in rs.split('\n') if s]
        #search for string #include "file.c"
        msearch = lambda s: re.search(r"\".*\.c\"", s)
        
        assert all(len(s) == 2 and msearch(s[1]) for s in rs), rs
        rs = [(os.path.dirname(s[0]), msearch(s[1]).group(0)[1:-1]) for s in rs]
        rs = set(os.path.join(fdir, fname) for fdir, fname in rs)
        return rs

    @staticmethod
    def analyze(results):
        #print 'world', results.results.subdir_pcs, results.subdir_pcs
        
        def print_set(s, name):
            if s:
                mlog.info("{} {}: {}".format(len(s), name, ', '.join(s))) 

        import time
        st = time.time()

        # find all subdirs with source in them
        import glob
        
        unit_files = (results.compilation_units | results.library_units |
                      results.hostprog_units | results.unconfigurable_units)
        
        subdirs = set(os.path.dirname(f) for f in unit_files)
        all_c_files = set()
        for d in subdirs:
            all_c_files.update([os.path.normpath(x) for x in glob.glob(os.path.join(d, "*.c"))])

        assert all(os.path.isfile(f) for f in all_c_files), all_c_files

        # find all compilation units without a corresponding .c file
        unmatched_units = set()
        asm_compilation_units = set()
        for unit in results.compilation_units:            
            S_file = FileAnalysis.otoS(unit)
            if os.path.isfile(S_file):
                asm_compilation_units.add(S_file)
            else:
                c_file = FileAnalysis.otoc(unit)
                if not os.path.isfile(c_file):
                    unmatched_units.add(c_file)

        #assert not unmatched_units, unmatched_units
        #assert not asm_compilation_units, asm_compilation_units

        # get source files that include c files
        included_c_files = set()
        out, _ = CM.vcmd(r'find . -name "*.[c|h]" | xargs grep -H "^#.*include.*\.c[\">]"')
        for line in out.split("\n"):
          parts = line.split(":", 1)
          if parts is not None:
            infile = parts[0]
            if len(parts) > 1:
              m = re.search(r"\".*\.c\"", parts[1])
              if m is not None:
                included = m.group(0)[1:-1]
                included = os.path.join(os.path.dirname(infile), included)
                included = os.path.relpath(os.path.realpath(included))
                included_c_files.add(included)

        # only need the files in the current source subtree
        included_c_files.intersection_update(all_c_files)

        # filtering
        # find all asm-offsets.c files, for these are compiled by the root
        # Kbuild file into offsets
        re_offsets_file = re.compile(r'arch/[^/]+/kernel/asm-offsets\.c')
        offsets_files = [f for f in all_c_files if re_offsets_file.match(x)]
        assert not offsets_files, offsets_files

        # find all .c files without a corresponding compilation unit, starting
        # with all c files
        unidentified_c_files = set(all_c_files)
        
        cu_c_files = [FileAnalysis.otoc(f) for f in results.compilation_units] #cu
        lib_c_files = [FileAnalysis.otoc(filename) for filename in results.library_units] #lib cu
        hostprog_c_files = [FileAnalysis.hostprog_otoc(f)  for f in results.hostprog_units]
        unconfig_c_files = [FileAnalysis.hostprog_otoc(f) for f in results.unconfigurable_units]
        extra_targets_c_files = [FileAnalysis.hostprog_otoc(f) for f in results.extra_targets]   
        used_c_files = set(cu_c_files + lib_c_files + hostprog_c_files + unconfig_c_files +
                           extra_targets_c_files + list(results.clean_files) + list(offsets_files))

        unidentified_c_files -= used_c_files

        # removing other c files
        # separate out .c files from the staging directory
        unidentified_staging_c_files = [f for f in unidentified_c_files 
                                        if "drivers/staging" in os.path.dirname(x)]
        unidentified_c_files -= set(unidentified_staging_c_files)

        # separate out .c files with "skeleton" in their name.  this is a
        # heuristic to find example drivers that aren't actually compiled.

        unidentified_skeleton_c_files = [f for f in unidentified_c_files 
                                         if "skeleton" in os.path.basename(x)]
        unidentified_c_files -= set(unidentified_skeleton_c_files)


        #remove .c files that aren't compilation units, because they are
        #included in other c files
        unidentified_c_files -= included_c_files

        # separate out generators heuristically.  look for "mk" or
        # "gen[^a-zA-Z]" in their name

        # perhaps we can find generators a better way, e.g., by looking for
        # the c file in the makefile

        # look for unexpanded variables or function calls
        re_unexpanded = re.compile(r'.*\$\(.*\).*')
        files = (results.compilation_units | results.library_units | results.hostprog_units |
                 results.unconfigurable_units | results.extra_targets | results.clean_files)
        unexpanded_units = set(FileAnalysis.otoc(f) for f in files if re_unexpanded.match(f))
        #remove files with unexpanded var names
        unmatched_units -= unexpanded_units

        # remove c files specified in the clean-files and in targets, since
        # these can be auto-generated c files
        import fnmatch        
        generated_c_files = set()
        for c in (results.clean_files | results.c_file_targets):
            pattern = re.compile(fnmatch.translate(c))
            for filename in unmatched_units:
                if pattern.match(filename):
                    generated_c_files.add(filename)
        unmatched_units.difference_update(generated_c_files)

        # # cache the list of working kbuild files
        # if args.excludes_file != None:
        #   with open(args.excludes_file, "w") as f:
        #     pickle.dump(excludes, f, 0)

        #print_set(toplevel_dirs, "toplevel_dirs")  # list of directories started from
        print_set(all_c_files, "all_c_files")  # all .c files in used and visited subdirs
        print_set(asm_compilation_units, "asm_compilation_units")  # compilation units with a .S file
        #print_set(subdirs, "subdirectory")  # subdirectory visited by kbuild
        #print_set(used_subdirectory, "used_subdirectory")  # subdirs containing compilation units
        print_set(results.compilation_units, "compilation_units")  # compilation units referenced by kbuild
        print_set(results.composites, "composites")  # compilation units that are composites
        print_set(results.library_units, "library_units")  # library units referenced by kbuild
        print_set(results.hostprog_units, "hostprog_units")
        print_set(results.unconfigurable_units, "unconfigurable_units")
        print_set(results.extra_targets, "extra_targets")
        print_set(results.clean_files, "clean_files")
        print_set(results.c_file_targets, "c_file_targets")
        print_set(generated_c_files, "generated_c_files")
        print_set(unmatched_units, "unmatched_units")
        print_set(included_c_files, "included_c_files")
        print_set(offsets_files, "offsets_files")
        print_set(unidentified_c_files, "unidentified_c_files")
        #print_set(unidentified_staging_c_files, "unidentified_staging_c_files")
        #print_set(unidentified_skeleton_c_files, "unidentified_skeleton_c_files")
        #print_set(unexpanded_compilation_units, "unexpanded_compilation_units")
        #print_set(unexpanded_library_units, "unexpanded_library_units")
        #print_set(unexpanded_hostprog_units, "unexpanded_hostprog_units")
        #print_set(unexpanded_unconfigurable_units, "unexpanded_unconfigurable_units")
        #print_set(unexpanded_extra_targets, "unexpanded_extra_targets")
        #print_set(unexpanded_subdirs, "unexpanded_subdirs")
        #print_set(broken, "broken")
        mlog.info("time: {}".format(time.time() - st))
    

# class GeneralAnalysis:
#     def __init__(self, path):
#         self.makefiles = path
#         # self.makefiles = set([os.path.abspath(f) for f in path
#         #                     if os.path.isfile(f) or os.path.isdir(f)])
#         mlog.info("analyzing {} files/dirs".format(len(self.makefiles)))

#     def run(self):
        
#         myrun = Run()
#         myrun.run(self.makefiles)
#         self.results = myrun.results

#     @property
#     def subdirectories(self):
#         return frozenset(self.results.subdirs)

#     @property
#     def compilation_units(self):
#         return frozenset(self.results.compilation_units)

#     @property
#     def composites(self):
#         return frozenset(self.results.composites)
    
#     @property
#     def library_units(self):
#         return frozenset(self.results.library_units)

#     @property
#     def hostprog_composites(self):
#         assert not self.results.hostprog_composites, self.results.hostprog_composites        
#         return frozenset(self.results.hostprog_composites)
#     @property
#     def hostprog_units(self):
#         return frozenset(self.results.hostprog_units)

#     @property    
#     def unconfigurable_units(self):
#         #/home/tnguyen/Dropbox/git/kmax-dev/tests/kbuild/missing_computed_variable
#         return frozenset(self.results.unconfigurable_units)
    
#     @property
#     def extra_targets(self):
#         assert not self.results.extra_targets, self.results.extra_targets
#         return frozenset(self.results.extra_targets)

#     @property
#     def clean_files(self):
#         return frozenset(self.results.clean_files)

#     @property
#     def c_file_targets(self):
#         assert not self.results.c_file_targets, self.results.c_file_targets
#         return frozenset(self.results.c_file_targets)

#     @property
#     def presence_conditions(self):
#         t = z3.Tactic('ctx-solver-simplify')
#         tseitin = z3.Tactic('tseitin-cnf')
#         presence_conditions = []
#         for path in self.results.presence_conditions.keys():
#             # presence_conditions.append((path, "{}".format(z3.simplify(self.results.presence_conditions[path]))))
#             presence_conditions.append((path, "{}".format(t(z3.simplify(self.results.presence_conditions[path])).as_expr())))
#             # presence_conditions.append((path, "{}".format(tseitin(t(z3.simplify(self.results.presence_conditions[path]))).as_expr())))
#         return frozenset(presence_conditions)

#     # @property
#     # def unit_pcs(self):
#     #     return self.results.unit_pcs

#     # @property
#     # def subdir_pcs(self):
#     #     return self.results.subdir_pcs
        

#     def analyze(self):

#     @classmethod
#     def chgext(cls, filename, f, t):
#         if f is None or filename.endswith(f):
#             return filename[:-2] + t
        
#     @classmethod
#     def hostprog_otoc(cls, filename):
#         """host programs don't have a .o extension, but their components
#         might if it's a composite"""
#         if not filename.endswith('.o'):
#             filename = filename + '.o'
#         return cls.chgext(filename, '.o', '.c')

#     @classmethod
#     def otoc(cls, filename):
#         return cls.chgext(filename, '.o', '.c')

#     @classmethod
#     def otoS(cls, filename):
#         return cls.chgext(filename, '.o', '.S')

#     @classmethod
#     def ctoo(cls, filename):
#         return chgext(filename, '.c', '.o')
    
#     @classmethod
#     def mkc(cls, filename):
#         return cls.chgext(filename, None, '.c')

#     @staticmethod
#     def get_all_c_files(path):
#         assert os.path.isdir(path)
#         cmd = 'find {} -type f | grep "\.c$" | sort | uniq'.format(path)
#         files, _ = CM.vcmd(cmd)
#         files = set([os.path.normpath(x.strip('\n'))
#                      for x in files.split()])
#         return files
        
#     @staticmethod
#     def get_included_c_files(path):
#         """get source files that include c files"""
        
#         path = os.path.abspath(path)
#         cmd = r'find {} -name "*.[c|h]" | xargs grep -H "^#.*include.*\.c[\">]"'.format(path)
#         rs, _  = CM.vcmd(cmd)
#         rs = [s.split(':', 1) for s in rs.split('\n') if s]
#         #search for string #include "file.c"
#         msearch = lambda s: re.search(r"\".*\.c\"", s)
        
#         assert all(len(s) == 2 and msearch(s[1]) for s in rs), rs
#         rs = [(os.path.dirname(s[0]), msearch(s[1]).group(0)[1:-1]) for s in rs]
#         rs = set(os.path.join(fdir, fname) for fdir, fname in rs)
#         return rs


# class Tests(GeneralAnalysis):
#     def __init__(self, topdir):
#         assert os.path.isdir(topdir), topdir
#         self.topdir = os.path.abspath(topdir)

#         files =[os.path.join(self.topdir, f) for f in os.listdir(self.topdir)]
#         self.makefiles = sorted(f for f in files if os.path.isfile(f))
#                 #for i in ~/Dropbox/git/kmax-dev/tests/kbuild/*; do echo "testing" $i &>> out ; kmax.py -g $i &>> out ; done
    

# class CaseStudy(GeneralAnalysis):
#     def __init__(self, topdir):
#         assert os.path.isdir(topdir)
        
#         self.topdir = os.path.abspath(topdir)
#         self.makefiles = set(os.path.join(self.topdir,d) for d in self.subdirs)
#         for d in list(self.makefiles):
#             if not (os.path.isdir(d) or os.path.isfile(d)):  #dir or makefile
#                 mlog.warn('{} not found .. removing'.format(d))
#                 self.makefiles.remove(d)

# class BusyboxCaseStudy(CaseStudy):
#     subdirs = set([
#       "applets/",
#       "archival/",
#       "archival/libarchive/",
#       "console-tools/",
#       "coreutils/",
#       "coreutils/libcoreutils/",
#       "debianutils/",
#       "klibc-utils/",
#       "e2fsprogs/",
#       "editors/",
#       "findutils/",
#       "init/",
#       "libbb/",
#       "libpwdgrp/",
#       "loginutils/",
#       "mailutils/",
#       "miscutils/",
#       "modutils/",
#       "networking/",
#       "networking/libiproute/",
#       "networking/udhcp/",
#       "printutils/",
#       "procps/",
#       "runit/",
#       "selinux/",
#       "shell/",
#       "sysklogd/",
#       "util-linux/",
#       "util-linux/volume_id/"
#     ])
    
#     additional_included = [
#       "archival/libarchive/unxz/xz_dec_bcj.c", # included in archival/libarchive/decompress_unxz.c
#       "archival/libarchive/unxz/xz_dec_lzma2.c",
#       "archival/libarchive/unxz/xz_dec_stream.c",

#       "archival/libarchive/bz/blocksort.c", # included in archival/bzip2.c
#       "archival/libarchive/bz/bzlib.c",
#       "archival/libarchive/bz/compress.c",
#       "archival/libarchive/bz/huffman.c",

#       "archival/libarchive/unxz/xz_dec_bcj.c", # included in archival/libarchive/decompress_unxz.c
#       "archival/libarchive/unxz/xz_dec_lzma2.c",
#       "archival/libarchive/unxz/xz_dec_stream.c",

#       "archival/libarchive/bz/blocksort.c", # included in archival/bzip2.c
#       "archival/libarchive/bz/bzlib.c",
#       "archival/libarchive/bz/compress.c",
#       "archival/libarchive/bz/huffman.c",

#       "archival/libarchive/unxz/xz_dec_bcj.c", # included in archival/libarchive/decompress_unxz.c
#       "archival/libarchive/unxz/xz_dec_lzma2.c",
#       "archival/libarchive/unxz/xz_dec_stream.c",

#       "archival/libarchive/bz/blocksort.c", # included in archival/bzip2.c
#       "archival/libarchive/bz/bzlib.c",
#       "archival/libarchive/bz/compress.c",
#       "archival/libarchive/bz/huffman.c"
#     ]
#     additional_hostprogs = [
#         "networking/ssl_helper/ssl_helper.c", # built by networking/ssl_helper/ssl_helper.sh
#         "networking/ssl_helper-wolfssl/ssl_helper.c" # built by networking/ssl_helper-wolfssl/ssl_helper.sh
#     ]

#     # explanations for busybox 1.25.0
#     explanations = {
#       "examples": [
#         "networking/httpd_ssi.c", # example program
#         "networking/httpd_indexcgi.c",

#         "shell/ash_test/printenv.c", # test program compiled and run in shell/ash_test/run-all
#         "shell/ash_test/recho.c",
#         "shell/ash_test/zecho.c",

#         "applets/individual.c", # example wrapper program

#         "networking/ntpd_simple.c", # for busybox 1_22_stable
#       ],

#       "unused": [
#         "util-linux/volume_id/unused_highpoint.c",  # unused source
#         "util-linux/volume_id/unused_hpfs.c",
#         "util-linux/volume_id/unused_isw_raid.c",
#         "util-linux/volume_id/unused_lsi_raid.c",
#         "util-linux/volume_id/unused_lvm.c",
#         "util-linux/volume_id/unused_mac.c",
#         "util-linux/volume_id/unused_minix.c",
#         "util-linux/volume_id/unused_msdos.c",
#         "util-linux/volume_id/unused_nvidia_raid.c",
#         "util-linux/volume_id/unused_promise_raid.c",
#         "util-linux/volume_id/unused_silicon_raid.c",
#         "util-linux/volume_id/unused_ufs.c",
#         "util-linux/volume_id/unused_via_raid.c",

#         "libbb/hash_md5prime.c", # commented out in libbb/Kbuild
#         "libbb/bb_strtod.c",

#         "networking/tc.c", # CONFIG_TC commented out of networking/Config.in
#       ],

#       "not built commented out": [
#         "editors/patch_bbox.c", # not built
#         "editors/patch_toybox.c",
#       ],
#     }
    
    
#     def analyze(self):
#         all_c_files = self.get_all_c_files(self.topdir)
#         len_all_c_files = len(all_c_files)

#         unit_files = self.compilation_units | self.library_units
#         unit_files = set(self.mkc(f) for f in unit_files)
#         all_c_files -= unit_files
        
#         additional_included = set(os.path.join(self.topdir, f)
#                                   for f in self.additional_included)
#         included_c_files = self.get_included_c_files(self.topdir)
#         all_included = included_c_files | additional_included
#         #print "included c files", len(all_included)
#         all_c_files -= all_included
        
#         prefix = os.path.join(self.topdir, "scripts/")
#         script_files = set(f for f in all_c_files if f.startswith(prefix))
#         #print "scripts", len(script_files)        
#         all_c_files -= script_files

#         hostprog_files = set(f + ".c" for f in self.hostprog_units)
#         additional_hostprogs = set(os.path.join(self.topdir, f)
#                                    for f in self.additional_hostprogs)
#         hostprog_files |= additional_hostprogs
#         #print "hostprogs", len(hostprog_files)
#         all_c_files -= hostprog_files

#         # busy 1_22_stable has an unused directory of source files
#         prefix = os.path.join(self.topdir, "e2fsprogs/old_e2fsprogs")
#         old_e2fsprogs = set(f for f in all_c_files if f.startswith(prefix))
#         #print "old e2fsprogs dir", len(old_e2fsprogs)
#         all_c_files -= old_e2fsprogs
        
#         for k in self.explanations:
#           print k, len(self.explanations[k])
#           fs = set(os.path.join(self.topdir, f) for f in self.explanations[k])
#           all_c_files -= fs
#           print len(all_c_files)
          
#         # This will be empty if kmax got all C files correctly.  The
#         # explanations needs to contain all C files guaranteed to not be part
#         # of the build.

#         #print some info
#         mlog.info("results00:\n{}".format(self.results))
        
#         d = [("all c files", len_all_c_files),
#              ("units", len(unit_files)),
#              ("included c files", len(all_included)),
#              ("scripts", len(script_files)),
#              ("hostprogs", len(hostprog_files)),
#              ("old e2fsprogs dir", len(old_e2fsprogs)),
#              ("unaccounted c files", len(all_c_files)),
#              ("unaccounted c files", ' '.join(all_c_files))
#         ]
#         s = ', '.join("{}: {}".format(k,v) for k,v in d)
#         mlog.info(s)
        

        
# class LinuxCaseStudy(GeneralAnalysis):
#     # subdirs = set([
#     #     'arch/x86/Makefile',
#     #     'ipc',
#     #     'arch/x86/pci',
#     #     'fs',
#     #     'kernel',
#     #     'arch/x86/video',
#     #     'arch/x86/power',
#     #     'security',
#     #     'sound',
#     #     'arch/x86',
#     #     'lib',
#     #     'mm',
#     #     'drivers',
#     #     'firmware',
#     #     'crypto',
#     #     'arch/x86/oprofile',
#     #     'init',
#     #     'usr',
#     #     'net',
#     #     'arch/x86/lib'
#     #     'block',
#     #     'arch/x86/math-emu'
#     # ])
    
#     def analyze(self):
#         pickle.dump((self.compilation_units,
#                      self.library_units,
#                      self.hostprog_units,
#                      self.unconfigurable_units,
#                      self.extra_targets,
#                      self.clean_files,
#                      self.c_file_targets,
#                      self.subdirectories,
#                      self.composites,
#                      self.presence_conditions), sys.stdout, 0)
        
#         # #print 'world', self.results.subdir_pcs, self.subdir_pcs
        
#         # def print_set(s, name):
#         #     if s:
#         #         mlog.info("{} {}: {}".format(len(s), name, ', '.join(s))) 

#         # import time
#         # st = time.time()
        
#         # # find all subdirs with source in them
#         # import glob
        
#         # unit_files = (self.compilation_units | self.library_units |
#         #               self.hostprog_units | self.unconfigurable_units)
        
#         # subdirs = set(os.path.dirname(f) for f in unit_files)
#         # all_c_files = set()
#         # for d in subdirs:
#         #     all_c_files.update([os.path.normpath(x) for x in glob.glob(os.path.join(d, "*.c"))])

#         # assert all(os.path.isfile(f) for f in all_c_files), all_c_files

#         # # find all compilation units without a corresponding .c file
#         # unmatched_units = set()
#         # asm_compilation_units = set()
#         # for unit in self.compilation_units:            
#         #     S_file = self.otoS(unit)
#         #     if os.path.isfile(S_file):
#         #         asm_compilation_units.add(S_file)
#         #     else:
#         #         c_file = self.otoc(unit)
#         #         if not os.path.isfile(c_file):
#         #             unmatched_units.add(c_file)

#         # #assert not unmatched_units, unmatched_units
#         # #assert not asm_compilation_units, asm_compilation_units


#         # # get source files that include c files
#         # included_c_files = set()
#         # out, _ = CM.vcmd(r'find . -name "*.[c|h]" | xargs grep -H "^#.*include.*\.c[\">]"')
#         # for line in out.split("\n"):
#         #   parts = line.split(":", 1)
#         #   if parts is not None:
#         #     infile = parts[0]
#         #     if len(parts) > 1:
#         #       m = re.search(r"\".*\.c\"", parts[1])
#         #       if m is not None:
#         #         included = m.group(0)[1:-1]
#         #         included = os.path.join(os.path.dirname(infile), included)
#         #         included = os.path.relpath(os.path.realpath(included))
#         #         included_c_files.add(included)

#         # # only need the files in the current source subtree
#         # included_c_files.intersection_update(all_c_files)

#         # # filtering
#         # # find all asm-offsets.c files, for these are compiled by the root
#         # # Kbuild file into offsets
#         # re_offsets_file = re.compile(r'arch/[^/]+/kernel/asm-offsets\.c')
#         # offsets_files = [f for f in all_c_files if re_offsets_file.match(x)]
#         # assert not offsets_files, offsets_files

#         # # find all .c files without a corresponding compilation unit, starting
#         # # with all c files
#         # unidentified_c_files = set(all_c_files)
        
#         # cu_c_files = [self.otoc(f) for f in self.compilation_units] #cu
#         # lib_c_files = [self.otoc(filename) for filename in self.library_units] #lib cu
#         # hostprog_c_files = [self.hostprog_otoc(f)  for f in self.hostprog_units]
#         # unconfig_c_files = [self.hostprog_otoc(f) for f in self.unconfigurable_units]
#         # extra_targets_c_files = [self.hostprog_otoc(f) for f in self.extra_targets]   
#         # used_c_files = set(cu_c_files + lib_c_files + hostprog_c_files + unconfig_c_files +
#         #                    extra_targets_c_files + list(self.clean_files) + list(offsets_files))

#         # unidentified_c_files -= used_c_files

#         # # removing other c files
#         # # separate out .c files from the staging directory
#         # unidentified_staging_c_files = [f for f in unidentified_c_files 
#         #                                 if "drivers/staging" in os.path.dirname(x)]
#         # unidentified_c_files -= set(unidentified_staging_c_files)

#         # # separate out .c files with "skeleton" in their name.  this is a
#         # # heuristic to find example drivers that aren't actually compiled.

#         # unidentified_skeleton_c_files = [f for f in unidentified_c_files 
#         #                                  if "skeleton" in os.path.basename(x)]
#         # unidentified_c_files -= set(unidentified_skeleton_c_files)


#         # #remove .c files that aren't compilation units, because they are
#         # #included in other c files
#         # unidentified_c_files -= included_c_files

#         # # separate out generators heuristically.  look for "mk" or
#         # # "gen[^a-zA-Z]" in their name

#         # # perhaps we can find generators a better way, e.g., by looking for
#         # # the c file in the makefile

#         # # look for unexpanded variables or function calls
#         # re_unexpanded = re.compile(r'.*\$\(.*\).*')
#         # files = (self.compilation_units | self.library_units | self.hostprog_units |
#         #          self.unconfigurable_units | self.extra_targets | self.clean_files)
#         # unexpanded_units = set(self.otoc(f) for f in files if re_unexpanded.match(f))
#         # #remove files with unexpanded var names
#         # unmatched_units -= unexpanded_units

#         # # remove c files specified in the clean-files and in targets, since
#         # # these can be auto-generated c files
#         # import fnmatch        
#         # generated_c_files = set()
#         # for c in (self.clean_files | self.c_file_targets):
#         #     pattern = re.compile(fnmatch.translate(c))
#         #     for filename in unmatched_units:
#         #         if pattern.match(filename):
#         #             generated_c_files.add(filename)
#         # unmatched_units.difference_update(generated_c_files)

#         # # # cache the list of working kbuild files
#         # # if args.excludes_file != None:
#         # #   with open(args.excludes_file, "w") as f:
#         # #     pickle.dump(excludes, f, 0)

#         # mlog.info("results01:\n{}".format(self.results))


        
#         # #print_set(toplevel_dirs, "toplevel_dirs")  # list of directories started from
#         # print_set(all_c_files, "all_c_files")  # all .c files in used and visited subdirs
#         # print_set(asm_compilation_units, "asm_compilation_units")  # compilation units with a .S file
#         # #print_set(subdirs, "subdirectory")  # subdirectory visited by kbuild
#         # #print_set(used_subdirectory, "used_subdirectory")  # subdirs containing compilation units
#         # print_set(self.compilation_units, "compilation_units")  # compilation units referenced by kbuild
#         # print_set(self.composites, "composites")  # compilation units that are composites
#         # print_set(self.library_units, "library_units")  # library units referenced by kbuild
#         # print_set(self.hostprog_units, "hostprog_units")
#         # print_set(self.unconfigurable_units, "unconfigurable_units")
#         # print_set(self.extra_targets, "extra_targets")
#         # print_set(self.clean_files, "clean_files")
#         # print_set(self.c_file_targets, "c_file_targets")
#         # print_set(generated_c_files, "generated_c_files")
#         # print_set(unmatched_units, "unmatched_units")
#         # print_set(included_c_files, "included_c_files")
#         # print_set(offsets_files, "offsets_files")
#         # print_set(unidentified_c_files, "unidentified_c_files")
#         # #print_set(unidentified_staging_c_files, "unidentified_staging_c_files")
#         # #print_set(unidentified_skeleton_c_files, "unidentified_skeleton_c_files")
#         # #print_set(unexpanded_compilation_units, "unexpanded_compilation_units")
#         # #print_set(unexpanded_library_units, "unexpanded_library_units")
#         # #print_set(unexpanded_hostprog_units, "unexpanded_hostprog_units")
#         # #print_set(unexpanded_unconfigurable_units, "unexpanded_unconfigurable_units")
#         # #print_set(unexpanded_extra_targets, "unexpanded_extra_targets")
#         # #print_set(unexpanded_subdirs, "unexpanded_subdirs")
#         # #print_set(broken, "broken")
#         # mlog.info("time: {}".format(time.time() - st))



import tempfile
import os.path
import itertools
import subprocess as sp
import operator
import inspect

import logging

def pause(s=None):
    try: #python2
        raw_input("Press any key to continue ..." if s is None else s)
    except NameError:
        input("Press any key to continue ..." if s is None else s)


def whoami():
    return inspect.stack()[1][3]

def vcmd(cmd, inp=None, shell=True):
    proc = sp.Popen(cmd,shell=shell,stdin=sp.PIPE,stdout=sp.PIPE,stderr=sp.PIPE)
    return proc.communicate(input=inp)
        
def vload(filename,mode='rb'):
    try:
        import cPickle as pickle
    except ImportError:  #Python3
        import pickle

    with open(filename,mode) as fh:
        pickler = pickle.Unpickler(fh)
        sobj = pickler.load()
    return sobj

def vsave(filename,sobj,mode='wb'):
    try:
        import cPickle as pickle
    except ImportError:  #Python3
        import pickle
        
    with open(filename,mode) as fh:
        pickler = pickle.Pickler(fh,-1)
        pickler.dump(sobj)

def vread(filename):
    with open(filename, 'r') as fh:
        return fh.read()

def iread(filename):
    """ return a generator """
    with open(filename, 'r') as fh:
        for line in fh:
            yield line

def strip_contents(lines, strip_c='#'):
    lines = (l.strip() for l in lines)
    lines = (l for l in lines if l)
    if strip_c:
        lines = (l for l in lines if not l.startswith(strip_c))
    return lines
    
def iread_strip(filename, strip_c='#'):
    """
    like iread but also strip out comments and empty line
    """
    return strip_contents(iread(filename), strip_c)


vmul = lambda l: reduce(operator.mul, l, 1)

getpath = lambda f: os.path.realpath(os.path.expanduser(f))
file_basename = lambda filename: os.path.splitext(filename)[0]

iflatten = lambda l: itertools.chain.from_iterable(l) #return a generator

# log utils
def getLogger(name, level):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(level)
    formatter = logging.Formatter("%(name)s:%(levelname)s:%(message)s")
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger

def getLogLevel(level):
    assert level in set(range(5))
    
    if level == 0:
        return logging.CRITICAL
    elif level == 1:
        return logging.ERROR
    elif level == 2:
        return logging.WARNING
    elif level == 3:
        return logging.INFO
    else:
        return logging.DEBUG




import pickle
import pdb
trace = pdb.set_trace
import z3
import kmaxtools.vcommon as CM
from dd.autoref import BDD

import kmaxtools.settings
mlog = CM.getLogger(__name__, kmaxtools.settings.logger_level)

# warning: this is not well encapsulated.  multiple runs in same process may not work properly
bdd_lib = BDD()

# todo: implement with dd package
def bdd_one(): return bdd_lib.true
def bdd_zero(): return bdd_lib.false
def bdd_ithvar(i):
    bdd_lib.add_var(i)
    return bdd_lib.var(i)
def bdd_init(): pass
def bdd_destroy(): pass

def conj(a, b): return None if a is None or b is None else a & b
def disj(a, b): return None if a is None or b is None else a | b
def neg(a): return None if a is None else ~a
def isbddfalse(b): return b == bdd_lib.false

def bdd_solutions(b):
    d = bdd_lib.pick(b)
    n = bdd_lib.count(b)
    print(n)
    expression = []
    solutions = [d for d in bdd_lib.pick_iter(b)]
    return solutions

class CondDef(tuple):
    def __new__(cls, cond, zcond, mdef):
        return super(CondDef, cls).__new__(cls, (cond, zcond, mdef))
    
    def __init__(self, cond, zcond, mdef):
        # assert isinstance(cond, pycudd.DdNode), cond
        assert z3.is_expr(zcond)
        assert mdef is None or isinstance(mdef, str), mdef  #CONFIG_A, 'y', 'm'
        self.cond = cond
        self.zcond = zcond
        self.mdef = mdef

    def __str__(self, printCond=None):
        if not printCond:
            return "{}".format(self.mdef)
        else:
            return "({}, {}, {})".format(self.cond, self.zcond, self.mdef)


class Multiverse(list):
    def __init__(self, ls=[]):
        #assert allisinstance(cd, CondDef) for cd in ls), ls
        
        list.__init__(self, ls)

    def __str__(self, printCond=None):
        return "CondDefs([{}])".format(
            ', '.join([p.__str__(printCond) for p in self]))
    
    def dedup(self):
        uniqs = set(cd.mdef for cd in self)
        if len(uniqs) == len(self):
            return self

        cache = {}
        for cond, zcond, val in self:
            if val in cache:
                c, zc = cache[val]
                cache[val] = (kmaxtools.alg.disj(c, cond), z3.Or(zc, zcond)) #disj
            else:
                cache[val] = (cond, zcond)

        mv = Multiverse([CondDef(c, zc, v)
                         for v, (c, zc)  in cache.iteritems()])
        assert mv
        return mv

class VarEntry(tuple):
    RECURSIVE = "RECURSIVE"
    SIMPLE = "SIMPLE"
    
    def __new__(cls, val, cond, zcond, flavor):
        return super(VarEntry, cls).__new__(cls, (val, cond, zcond, flavor))
    
    def __init__(self, val, cond, zcond, flavor):
        assert val is None or isinstance(val, str), val
        # assert isinstance(cond, pycudd.DdNode), cond
        assert z3.is_expr(zcond), zcond
        assert flavor in set({VarEntry.RECURSIVE, VarEntry.SIMPLE}), flavor

        self.val = val.strip() if isinstance(val, str) else val
        self.cond = cond
        self.zcond = z3.simplify(zcond)
        self.flavor = flavor

    def __str__(self, printCond=None):
        ss = [self.val, self.flavor]

        if printCond:
            ss.append(str(self.cond))
            
        ss.append(self.zcond)
            
        return "; ".join(map(str,ss))

    @property
    def condDef(self):
        return CondDef(self.cond, self.zcond, self.val)

class BoolVar(tuple):
    def __new__(cls, bdd, zbdd, idx):
        # assert isinstance(bdd, pycudd.DdNode), bdd
        assert idx >= 0, idx
        
        return super(BoolVar, cls).__new__(cls, (bdd, zbdd, idx))
    
    def __init__(self, bdd, zbdd, idx):
        assert z3.is_expr(zbdd), zbdd
        assert idx >= 0, idx
        
        self.bdd = bdd
        self.zbdd = zbdd
        self.idx = idx

    def __str__(self, printCond=None):
        ss = [self.idx, self.zbdd]
        if printCond:
            ss.append(str(self.bdd))
        return ", ".join(map(str,ss))


class Results:
    def __init__(self):
        self.subdirs = set()
        self.compilation_units = set()
        self.library_units = set()
        self.composites = set()
        self.hostprog_units = set()
        self.hostprog_composites = set()
        self.unconfigurable_units = set()
        self.extra_targets = set()
        self.clean_files = set()
        self.c_file_targets = set()
        # self.unit_pcs = set()
        # self.subdir_pcs = set()
        self.presence_conditions = {}

    def __str__(self, details=False):
        if kmaxtools.settings.output_smtlib2:
            z3_pcs = {}
            for filename in self.presence_conditions.keys():
                solver = z3.Solver()
                solver.add(self.presence_conditions[filename])
                z3_pcs[filename] = solver.to_smt2()
            return pickle.dumps(z3_pcs, 0).decode()
        else:
            f = lambda k, s: "{}: {}".format(k, ', '.join(s) if details else len(s))
            delim = '\n' if details else ', '
            ss = delim.join(f(k, s) for k, s in self.__dict__.iteritems() if s)
            if self.presence_conditions:
                ss += '\n{} presence conditions: \n{}'.format(len(self.presence_conditions), self.z3_str(self.presence_conditions))
            # if self.unit_pcs:
            #     ss += '\n{} unit pcs: \n{}'.format(len(self.unit_pcs), self.pc_str(self.unit_pcs))
            # if self.subdir_pcs:
            #     ss += '\n{} subdir pcs: \n{}'.format(len(self.subdir_pcs), self.pc_str(self.subdir_pcs))
            return ss
    
    def pc_str(self, s):
        return '\n'.join("{}. {}: {}, {}".format(i, v, f, z3.simplify(g))
                         for i, (v, f, g) in enumerate(s))

    def split_nested(self, s):
        # assume we've come in with "CONFIG_A)", i.e., no beginning
        # top-level lparen, but the matching rparen for it, because we
        # split on it in to_exp_step
        depth = 1
        operands = []
        operand = ""
        for c in s:
            if c=='(':
                depth += 1
                
            if c==')':
                depth -= 1

            if c==',' and depth==1 or c==')' and depth==0:
                operands.append(operand)
                # print operand
                operand = ""
            else:
                operand += c
        assert depth == 0  # we started without the top-level lparen, but with the top-level rparen, so we started on depth 1 and should exit on depth 0
        return operands
    
    def to_exp_step(self, s):
        # print s
        if s.startswith("And") or s.startswith("Or") or s.startswith("Not"):
            (opname, operands) = s.split('(', 1)
            operands = self.split_nested(operands)
            operands = [ self.to_exp_step(operand) for operand in operands ]
            # print opname, operands
            new_s = ""
            if opname=="And":
                new_s = "(" + " && ".join(operands) + ")"
            elif opname=="Or":
                new_s = "(" + " || ".join(operands) + ")"
            elif opname=="Not":
                assert len(operands) == 1
                new_s = "(! " + operands[0] + ")"
            # print "new_s", new_s
            return new_s
        if s=="True":
            return "1"
        else:
            return s
    
    def to_exp(self, s):
        s = str(s).replace('\n', '').replace(' ', '')
        return self.to_exp_step(s)

    def get_line_format(self, filename):
        if filename.endswith("/"):
            line_format = "subdir_pc {} {}"
        else:
            line_format = "unit_pc {} {}"

        return line_format
    
    def z3_str(self, path_conds):
        # print self.to_exp("And(CONFIG_RMMOD, Not(CONFIG_MODPROBE_SMALL))")
        # exit(1)
        # result = '\n'.join("unit_pc {} {}".format(i, self.to_exp(z3.simplify(s[i]))) + "\nunit_pc {} {}".format(i, z3.simplify(s[i]))
        #                  for i in s.keys())

        # # separate paths into files and subdirs, which end in forward slash
        # file_pcs = { path : cond for path, cond in path_conds.iteritems() if not path.endswith("/") }
        # subdir_pcs = { path : cond for path, cond in path_conds.iteritems() if path.endswith("/") }

        # # find full presence conditions for each file and conjoining the presence conditions for each path
        # full_pcs = {}

        # for path, cond in file_pcs.iteritems():
        #     subpath, basename = path.rsplit('/', 1)
        #     elems = subpath.rsplit('/')
        #     for i in reversed(range(0, 7)):
        #         subsubpath = '/'.join(elems[:-i])
        #         if len(subsubpath) == 0: subsubpath = "/"
        #         print subsubpath, subsubpath in subdir_pcs
        #         if subsubpath in subdir_pcs: print subsubpath, subdir_pcs[subsubpath]

        result = '\n'.join(self.get_line_format(filename).format(filename, self.to_exp(z3.simplify(path_conds[filename])))
                           for filename in path_conds.keys())
        return result



import sys
import os
import re
import argparse
import subprocess as sp
import tempfile

from pymake import parser, parserdata, data, functions
from collections import defaultdict

import pdb
trace = pdb.set_trace
import z3
import kmaxtools.vcommon as CM

from datastructures import CondDef, Multiverse, VarEntry, BoolVar, Results

import kmaxtools.settings
mlog = CM.getLogger(__name__, kmaxtools.settings.logger_level)

match_unexpanded_variables = re.compile(r'.*\$\(.*\).*')
def contains_unexpanded(s):
    """
    return for strings such as $(test)
    """
    assert s is None or isinstance(s, str), s

    return s is not None and match_unexpanded_variables.match(s)

# wrappers for symbolic boolean operations
def conj(a, b): return kmaxtools.datastructures.conj(a, b)
def disj(a, b): return kmaxtools.datastructures.disj(a, b)
def neg(a): return kmaxtools.datastructures.neg(a)
def bdd_one(): return kmaxtools.datastructures.bdd_one()
def bdd_zero(): return kmaxtools.datastructures.bdd_zero()
def bdd_ithvar(i): return kmaxtools.datastructures.bdd_ithvar(i)
def bdd_init(): kmaxtools.datastructures.bdd_init()
def bdd_destroy(): kmaxtools.datastructures.bdd_destroy()
def isbddfalse(b): return kmaxtools.datastructures.isbddfalse(b)

def zconj(a, b): return None if a is None or b is None else z3.simplify(z3.And(a, b))
def zdisj(a, b): return None if a is None or b is None else z3.simplify(z3.Or(a, b))
def zneg(a): return None if a is None else z3.simplify(z3.Not(a))


def isz3false(z):
  s = z3.Solver()
  s.add(z3.simplify(z))
  return z3.sat != s.check()

def isfalse(b, z):
    # return isz3false(z)
    return isbddfalse(b)

def isBooleanConfig(name):
    # if all_config_var_names is not None and not isNonbooleanConfig(name):
    #     return name in all_config_var_names
    # return False
    return False


def isNonbooleanConfig(name):
    # if nonboolean_config_var_names is not None:
    #     return name in nonboolean_config_var_names
    # return False
    return False

class ZSolver:
    T = z3.BoolVal(True)
    F = z3.BoolVal(False)        
    
    def __init__(self):
        self.solver = z3.Solver()

    def zcheck(self, f):
        self.solver.push()
        self.solver.add(f)
        ret = self.solver.check()
        self.solver.pop()
        return ret

class Kbuild:
    def __init__(self):
        bdd_init()
        self.zsolver = ZSolver()        

        # Boolean constants
        self.T = bdd_one()
        self.F = bdd_zero()

        self.variables = {}
        self.bvars = {}
        self.reverse_bvars = {}
        self.undefined_variables = set()
        # self.token_pc = {} # token presence conds, e.g., 'fork.o': True
        # self.unit_pc = {} # compilation unit presence conds
        # self.subdir_pc = {} # subdir presence conds
        # self.composite_pc = {} # composite presence conditions
        # variable equivalence classes for optimized append
        self.var_equiv_sets = {}

    def process_stmts(self, stmts, cond, zcond):
        """Find configurations in the given list of stmts under the
        given presence cond."""
        for s in stmts:
            if isinstance(s, parserdata.ConditionBlock):
                self.process_conditionblock(s, cond, zcond)
            elif isinstance(s, parserdata.SetVariable):
                self.process_setvariable(s, cond, zcond)
            elif isinstance(s, (parserdata.Rule, parserdata.StaticPatternRule)):
                self.process_rule(s, cond, zcond)
            elif (isinstance(s, parserdata.Include)):
                self.process_include(s, cond, zcond)
            else:
                mlog.warn("cannot parse ({})".format(s))

    def get_bvars(self, name):
        # Mapping of boolean variable names to BDD variable number.
        # Automatically increments the variable number.
        
        assert isinstance(name, str), name
        try:
            return self.bvars[name]
        except KeyError:
            idx = len(self.bvars)
            bdd = bdd_ithvar(idx)
            zbdd = z3.Bool(name.format(idx))
            bv = BoolVar(bdd, zbdd, idx)
            self.bvars[name] = bv
            self.reverse_bvars[idx] = name
            return bv
    
    def get_var_equiv(self, name):
        # get a new randomized variable name that is equivalent to the
        # given variable.  this also updates a structure that records
        # which variables are equivalent

        assert isinstance(name, str), name

        if name in self.var_equiv_sets:
            name_ = "{}_EQUIV{}".format(len(self.var_equiv_sets[name]), name)
            self.var_equiv_sets[name].add(name_)
            return name_
        else:
            self.var_equiv_sets[name] = set([name])
            return name
        
    def get_var_equiv_set(self, name):
        if name not in self.var_equiv_sets:
            self.var_equiv_sets[name] = set([name])
            
        return self.var_equiv_sets[name]

    def getSymbTable(self, printCond=None):

        f = lambda vs: '\n'.join("{}. {}".format(i + 1, v.__str__(printCond))
                                 for i, v in enumerate(vs))

        ss = [(name, [v for v in self.variables[name] if v.val])
              for name in self.variables]
        ss = ["var: {}:\n{}\n---------".format(name, f(vs))
              for name, vs in ss if vs]

        return '\n'.join(ss)
    
    def get_presence_conditions(self, vars, pcs, cond, zcond):
        names = set()
        for var in vars:
            if var in self.variables.keys():
                names = names.union(self.get_var_equiv_set(var))
        while len(names) > 0:
            name = names.pop()
            # print name
            # print self.variables[name]
            for value, bdd_condition, z3_condition, flavor in self.variables[name]:
                # print "  ---"
                # print "  VALUE:", value
                # print "  BDD:", self.bdd_to_str(bdd_condition)
                # print "  Z3:", z3_condition
                # print "  FLAVOR:", flavor
                tokens = value.split()
                for token in tokens:
                    and_cond = conj(cond, bdd_condition)
                    and_zcond = zconj(zcond, z3_condition)
                    if token not in pcs.keys():
                        pcs[token] = and_zcond
                    else:
                        pcs[token] = zdisj(pcs[token], and_zcond)
                    if token.endswith(".o"): # and unit_name not in compilation_units:
                        if (token[:-2] + "-objs") in self.variables or \
                            (token[:-2] + "-y") in self.variables:
                            # scripts/Makefile.build use the -objs and -y
                            # suffix to define composites $($(subst
                            # $(obj)/,,$(@:.o=-objs))) $($(subst
                            # $(obj)/,,$(@:.o=-y)))), $^)
                            composite_variable1 = token[:-2] + "-objs"
                            composite_variable2 = token[:-2] + "-y"
                            self.get_presence_conditions([ composite_variable1, composite_variable2 ], pcs, and_cond, and_zcond)

    def add_definitions(self, defines):
        if not defines:
            return
        
        for define in defines:
            name, value = define.split("=")
            self.add_var(name, self.T, ZSolver.T, '=', value)

    def get_defined(self, variable, expected):
        # variable_name = "defined(" + variable + ")"
        # todo implement for tristate flag also, i.e., =y or =m and !=n
        variable_name = variable
        bdd = self.get_bvars(variable_name).bdd
        zbdd = self.get_bvars(variable_name).zbdd
        
        if expected:
            return bdd, zbdd
        else:
            return neg(bdd), z3.Not(zbdd)

    def process_variableref(self, name):
        if name not in self.variables and name == 'BITS':
            # TODO get real entry from top-level makefiles
            bv32 = self.get_bvars("BITS=32")
            bv64 = self.get_bvars("BITS=64")
            return Multiverse([ CondDef(bv32.bdd, bv32.zbdd, "32"),
                                CondDef(bv64.bdd, bv64.zbdd, "64") ])
        
        # elif name == 'ARCH':
        #     # TODO user-defined
        #     return Multiverse([ (self.T, "x86") ])
        elif name == "CONFIG_WORD_SIZE":

            raise NotImplementedError
            # TODO get defaults from Kconfig files
            return Multiverse([ (self.bvars["CONFIG_WORD_SIZE=32"].bdd, "32"),
                                (self.bvars["CONFIG_WORD_SIZE=64"].bdd, "64") ])

        elif name not in self.variables and name == "MMU":
            raise NotImplementedError
        
            # TODO get globals from arch Makefiles
            is_defined = self.get_defined("CONFIG_MMU", True)
            not_defined = self.get_defined("CONFIG_MMU", False)

            return Multiverse([ (is_defined, ''),
                                (not_defined, '-nommu') ])

        elif name.startswith("CONFIG_"):
            if kmaxtools.settings.do_boolean_configs:
                #varbdd = self.bvars[name].bdd
                v = self.get_bvars(name).bdd
                zv = self.get_bvars(name).zbdd
            
                return Multiverse([ CondDef(v, zv, 'y'),
                                    CondDef(neg(v), z3.Not(zv), None) ])
            else:
                # TODO don't use 'm' for truly boolean config vars
                equals_y = self.get_bvars(name + "=y").bdd
                zequals_y = self.get_bvars(name + "=y").zbdd

                equals_m = self.get_bvars(name + "=m").bdd
                zequals_m = self.get_bvars(name + "=m").zbdd

                defined, zdefined = self.get_defined(name, True)
                is_defined_y = conj(defined, conj(equals_y, neg(equals_m)))
                zis_defined_y = zconj(zdefined, zconj(zequals_y, z3.Not(zequals_m)))


                is_defined_m = conj(defined, conj(equals_m, neg(equals_y)))
                zis_defined_m = zconj(zdefined, zconj(zequals_m, z3.Not(zequals_y)))

                notdefined, znotdefined = self.get_defined(name, False)
                not_defined = disj(notdefined, conj(neg(is_defined_y), neg(is_defined_m)))
                znot_defined = zdisj(znotdefined, zconj(z3.Not(zis_defined_y), z3.Not(zis_defined_m)))   


                return Multiverse([ CondDef(is_defined_y, zis_defined_y, 'y'),
                                    CondDef(is_defined_m, zis_defined_m, 'm'),
                                    CondDef(not_defined, znot_defined, None) ])
        # elif (name.startswith("CONFIG_")) and not isNonbooleanConfig(name):
        #     return Multiverse([ (self.T, '') ])
        else:
            if name in self.undefined_variables:
                return Multiverse([v.condDef for v in self.variables[name]])
            
            elif name not in self.variables and not isNonbooleanConfig(name):
                # Leave undefined variables unexpanded
                self.undefined_variables.add(name)
                self.variables[name] = [VarEntry("$(%s)" % (name),
                                                 self.T, ZSolver.T,
                                                 VarEntry.RECURSIVE)]

                mlog.warn("Undefined variable expansion: {}".format(name))
                return Multiverse([v.condDef for v in self.variables[name]])
            
            else:
                expansions = []
                equivs = self.get_var_equiv_set(name)
                for equiv_name in equivs:
                    for v in self.variables[equiv_name]:
                        if v.val:
                            expansions = expansions + self.expand_and_flatten(v.val, v.cond, v.zcond)
                        else:
                            expansions.append(v.condDef)

                return Multiverse(expansions)

    def process_function(self, function):
        """Evaluate variable expansion or built-in function and return
        either a single string or a list of (condition, string) pairs."""
        if isinstance(function, functions.VariableRef):
            return self.process_fun_VariableRef(function)
        elif isinstance(function, functions.SubstFunction):
            return self.process_fun_SubstFunction(function)
        elif isinstance(function, functions.SubstitutionRef):
            return self.process_fun_SubstitutionRef(function)
        elif isinstance(function, functions.IfFunction):
            return self.process_fun_IfFunction(function)
        elif isinstance(function, functions.FilteroutFunction):
            return self.process_fun_FilteroutFunction(function)
        elif isinstance(function, functions.PatSubstFunction):
            return self.process_fun_PatSubstFunction(function)
        elif isinstance(function, functions.SortFunction):
            return Multiverse(self.mk_Multiverse(self.process_expansion(function._arguments[0])))
        elif isinstance(function, functions.AddPrefixFunction):
            return self.process_fun_AddPrefixFunction(function)
        else:
            mlog.warn("default on function: {}".format(function))
            return self.mk_Multiverse(function.to_source())

    def process_fun_VariableRef(self, function):
        name = self.mk_Multiverse(self.process_expansion(function.vname))
        expanded_names = []
        for name_cond, name_zcond, name_value  in name:
            expanded_name = self.process_variableref(name_value)
            assert isinstance(expanded_name, Multiverse), expanded_name

            for v in expanded_name:
                expanded_names.append(
                    CondDef(conj(name_cond, v.cond), zconj(name_zcond, v.zcond), v.mdef)
                )
        return Multiverse(expanded_names)

    def process_fun_SubstFunction(self, function):
        from_vals = self.mk_Multiverse(self.process_expansion(function._arguments[0]))
        to_vals = self.mk_Multiverse(self.process_expansion(function._arguments[1]))
        in_vals = self.mk_Multiverse(self.process_expansion(function._arguments[2]))

        # Hoist conditionals around the function by getting all
        # combinations of arguments
        hoisted_results = []
        for (sc, szc, s), (rc, rzc, r), (dc, dzc, d) in zip(from_vals, to_vals, in_vals):
            instance_cond = conj(sc, conj(rc, dc))
            instance_zcond = z3.simplify(z3.And(szc, rzc, dzc))
            if not isfalse(instance_cond, instance_zcond):
                if r is None: r = ""  # Fixes bug in net/l2tp/Makefile
                instance_result = None if d is None else d.replace(s, r)
                hoisted_results.append(CondDef(instance_cond, instance_zcond, instance_result))

        return Multiverse(hoisted_results)
        
    def process_fun_IfFunction(self, function):
        cond_part = self.mk_Multiverse(self.process_expansion(function._arguments[0]))
        then_part = self.mk_Multiverse(self.process_expansion(function._arguments[1]))
        then_cond = self.F
        then_zcond = ZSolver.F
        else_cond = self.F
        else_zcond = ZSolver.F
        for cond, zcond, value in cond_part:
            if value:
                then_cond = disj(then_cond, cond)
                then_zcond = zdisj(then_zcond, zcond)
            else:
                else_cond = disj(then_cond, cond)
                else_zcond = zdisj(then_zcond, zcond)

        expansions = []

        for cond, zcond, value in then_part:
            cond = conj(then_cond, cond)
            zcond = zconj(then_zcond, zcond)
            if not isfalse(cond, zcond):
                expansions.append(CondDef(cond, zcond, value))

        if len(function._arguments) > 2:
            else_part = self.mk_Multiverse(self.process_expansion(function._arguments[2]))
            for cond, zcond, value in else_part:
                cond = conj(else_cond, cond)
                zcond = zconj(else_zcond, zcond)

                if not isfalse(cond, zcond):
                    expansions.append(CondDef(cond, zcond, value))

        return Multiverse(expansions)

    def process_fun_FilteroutFunction(self, function):
        from_vals = self.mk_Multiverse(self.process_expansion(function._arguments[0]))
        in_vals = self.mk_Multiverse(self.process_expansion(function._arguments[1]))

        # Hoist conditionals around the function by getting all
        # combinations of arguments
        hoisted_args = tuple((s, d) for s in from_vals for d in in_vals)
        hoisted_results = []
        # Compute the function for each combination of arguments
        for (c1, zc1, s), (c2, zc2, d) in hoisted_args:
            instance_cond = conj(c1, c2)
            instance_zcond = zconj(zc1, zc2)

            if not isfalse(instance_cond, instance_zcond):
                if d is None:
                    instance_result = None
                else:
                    instance_result = " ".join(d_token for d_token in d.split() if d_token != s)

                cd = CondDef(instance_cond, instance_zcond, instance_result)
                hoisted_results.append(cd)

        return Multiverse(hoisted_results)

    def process_fun_PatSubstFunction(self, function):
        from_vals = self.mk_Multiverse(self.process_expansion(function._arguments[0]))
        to_vals = self.mk_Multiverse(self.process_expansion(function._arguments[1]))
        in_vals = self.mk_Multiverse(self.process_expansion(function._arguments[2]))

        # Hoist conditionals around the function by getting all
        # combinations of arguments
        hoisted_args = tuple((s, r, d)
                             for s in from_vals
                             for r in to_vals
                             for d in in_vals)
        hoisted_results = []
        # Compute the function for each combination of arguments
        for (c1, zc1, s), (c2, zc2, r), (c3, zc3, d) in hoisted_args:
            instance_cond = conj(c1, conj(c2, c3))
            instance_zcond = z3.simplify(z3.And(zc1, zc2, zc3))
            
            if not isfalse(instance_cond, instance_zcond):
                if r == None: r = ""  # Fixes bug in net/l2tp/Makefile
                pattern = "^" + s.replace(r"%", r"(.*)", 1) + "$"
                replacement = r.replace(r"%", r"\1", 1)
                if d is None:
                    instance_result = None
                else:
                    instance_result = " ".join(re.sub(pattern, replacement, d_token)
                                               for d_token in d.split())
                cd = CondDef(instance_cond, instance_zcond, instance_result)
                hoisted_results.append(cd)
                
        return Multiverse(hoisted_results)

    def process_fun_AddPrefixFunction(self, function):
        prefixes = self.mk_Multiverse(self.process_expansion(function._arguments[0]))
        token_strings = self.mk_Multiverse(self.process_expansion(function._arguments[1]))

        hoisted_results = []
        for (prefix_cond, prefix_zcond, prefix) in prefixes:
            for (tokens_cond, tokens_zcond, token_string) in token_strings:
                resulting_cond = conj(prefix_cond, tokens_cond)
                resulting_zcond = zconj(prefix_zcond, tokens_zcond)

                if not isfalse(resulting_cond, resulting_zcond):
                    # append prefix to each token in the token_string
                    if token_string is None:
                        prefixed_tokens = ""                            
                    else:
                        prefixed_tokens = " ".join(prefix + token
                                                   for token in token_string.split())
                    hoisted_results.append(CondDef(resulting_cond, resulting_zcond, prefixed_tokens))

        return Multiverse(hoisted_results)
        
    def process_fun_SubstitutionRef(self, function):
        raise NotImplementedError

        from_values = self.mk_Multiverse(self.process_expansion(function.substfrom))
        to_values = self.mk_Multiverse(self.process_expansion(function.substto))
        name = self.mk_Multiverse(self.process_expansion(function.vname))

        # first expand the variable
        in_values = []
        for name_cond, name_value in name:
            expanded_name = self.process_variableref(name_value)
            for (expanded_cond, expanded_value) in expanded_name:
                in_values.append( (conj(name_cond, expanded_cond), expanded_value) )

        # then do patsubst
        hoisted_arguments = tuple((s, r, d)
                                  for s in from_values
                                  for r in to_values
                                  for d in in_values)

        hoisted_results = []
        for (c1, s), (c2, r), (c3, d) in hoisted_arguments:
            instance_condition = conj(c1, conj(c2, c3))
            if instance_condition != self.F:
                if r == None: r = ""  # Fixes bug in net/l2tp/Makefile
                if r"%" not in s:
                    # without a %, use a % to implement replacing
                    # the end of the string
                    s = r"%" + s
                    r = r"%" + r
                pattern = "^" + s.replace(r"%", r"(.*)", 1) + "$"
                replacement = r.replace(r"%", r"\1", 1)
                if d is None:
                    instance_result = None
                else:
                    instance_result = " ".join([re.sub(pattern, replacement, d_token) for d_token in d.split()])
                hoisted_results.append((instance_condition, instance_result))

        return Multiverse(hoisted_results)
        
    def mk_Multiverse(self, expansion):
        assert isinstance(expansion, (str, Multiverse)), expansion
        
        if isinstance(expansion, str):
            return Multiverse([CondDef(self.T, ZSolver.T, expansion)])
        else:
            return expansion

    def process_element(self, element, isfunc):
        if isinstance(element, str):
            return element
        elif isfunc:
            ret = self.process_function(element)
            return ret
        else:
            ret = self.process_expansion(element)
            return ret

    def hoist(self, expansion):
        """Hoists a list of expansions, strings, and Multiverses.
        return a Multiverse
        """
        #trace()
        hoisted = [(self.T, ZSolver.T, [])]
        for element in expansion:
            if isinstance(element, Multiverse):
                newlist = []
                for subcondition, zsubcondition, subverse in element:
                    for condition, zcondition, verse in hoisted:
                        newcondition = conj(condition, subcondition)
                        newzcondition = zconj(zcondition, zsubcondition)

                        # filter infeasible combinations
                        if not isfalse(newcondition, newzcondition):
                            newverse = list(verse)
                            if isinstance(subverse, list):
                                newverse += subverse
                            else:
                                newverse.append(subverse)
                            newlist.append((newcondition, newzcondition, newverse))
                hoisted = newlist
            else:
                for condition, zcondition, verse in hoisted:
                    verse.append(element)

        multiverse = Multiverse([CondDef(condition, zcondition, self.join_values(verse))
                                 for condition, zcondition, verse in hoisted] )

        multiverse = multiverse.dedup()
        return multiverse

    def process_expansion(self, expansion):
        """Expand variables in expansion, hoisting multiply-defined ones

        @param expansion is a pymake Expansion object

        @return either a single string or a Multiverse, i.e., list of (condition, string)
        pairs."""

        if isinstance(expansion, data.StringExpansion):
            return expansion.s
        elif isinstance(expansion, data.Expansion):
            rs = [self.process_element(element, isfunc)
                  for element, isfunc in expansion]
            mv = self.hoist(rs)
            assert isinstance(mv, Multiverse), mv
            # print("process_expansion", mv)
            return mv
        else:
            mlog.error("Unsupported BaseExpansion subtype: {}".format(expansion))
            exit(1)

    def process_conditionblock(self, block, presence_cond, presence_zcond):
        """Find a satisfying configuration for each branch of the
        conditional block."""
        # See if the block has an else branch.  Assumes no "else if".
        if len(block) == 1: has_else = False
        elif len(block) == 2: has_else = True
        else:
            mlog.warn("unsupported conditional block: {}".format(block))
            return

        # Process first branch
        cond, stmts = block[0]  # condition is a Condition object
        first_branch_cond = None
        if isinstance(cond, parserdata.IfdefCondition):  # ifdef
            # TODO only care if condition.exp.variable_references contains
            # multiply-defined macros
            expansion = self.process_expansion(cond.exp)
            if isinstance(expansion, str):
                first_branch_cond, first_branch_zcond = self.get_defined(expansion, cond.expected)
            else:
                #trace()
                cds = [cd for cd in expansion if cd.mdef is not None]
                hoisted_cond = reduce(disj, [cd.cond for cd in cds])
                hoisted_zcond = z3.simplify(z3.Or([cd.zcond for cd in cds]))

                first_branch_cond = hoisted_cond
                first_branch_zcond = hoisted_zcond

            first_branch_cond = conj(presence_cond, first_branch_cond)
            first_branch_zcond = zconj(presence_zcond, first_branch_zcond)

            else_branch_cond = neg(first_branch_cond)
            else_branch_zcond = z3.Not(first_branch_zcond)

        elif isinstance(cond, parserdata.EqCondition):  # ifeq
            exp1 = self.mk_Multiverse(self.process_expansion(cond.exp1))
            exp2 = self.mk_Multiverse(self.process_expansion(cond.exp2))

            # Hoist multiple expansions around equality operation
            hoisted_cond = self.F
            hoisted_zcond = ZSolver.F
            
            hoisted_else = self.F
            hoisted_zelse = ZSolver.F

            for cd1 in exp1:
                v1 = cd1.mdef if cd1.mdef else ""                
                for cd2 in exp2:
                    v2 = cd2.mdef if cd2.mdef else ""
                    term_cond = conj(cd1.cond, cd2.cond)
                    term_zcond = zconj(cd1.zcond, cd2.zcond)

                    #TODO: check term_zcond == term_cond
                    
                    if not isfalse(term_cond, term_zcond) and (v1 == v2) == cond.expected:
                        hoisted_cond = disj(hoisted_cond, term_cond)
                        hoisted_zcond = zdisj(hoisted_zcond, term_zcond)
                        
                    elif not isfalse(term_cond, term_zcond):
                        hoisted_else = disj(hoisted_else, term_cond)
                        hoisted_zelse = zdisj(hoisted_zelse, term_zcond)
                        
                    if contains_unexpanded(v1) or contains_unexpanded(v2):
                        # this preserves configurations where we
                        # didn't have values for a config option
                        v = self.get_bvars("{}={}".format(v1, v2))
                        bdd = v.bdd
                        zbdd = v.zbdd
                        hoisted_cond = disj(hoisted_cond, bdd)
                        hoisted_zcond = zdisj(hoisted_zcond, zbdd)
                        
                        hoisted_else = disj(hoisted_else, neg(bdd))
                        hoisted_zelse = zdisj(hoisted_zelse, z3.Not(zbdd))

                first_branch_cond = conj(presence_cond, hoisted_cond)
                first_branch_zcond = zconj(presence_zcond, hoisted_zcond)      
                else_branch_cond = conj(presence_cond, hoisted_else)
                else_branch_zcond = zconj(presence_zcond, hoisted_zelse)
                
        else:
            mlog.error("unsupported conditional branch: {}".format(cond))
            exit(1)

        assert first_branch_zcond is not None, \
            "Could not get if branch cond {}".format(first_branch_zcond)
        
        # Enter first branch
        # trace()
        # print("process_conditionblock", z3.simplify(first_branch_zcond))
        self.process_stmts(stmts, first_branch_cond, first_branch_zcond)

        if not has_else:
            return

        # Process the else branch
        cond, stmts = block[1]
        self.process_stmts(stmts, else_branch_cond, else_branch_zcond)  # Enter else branch

    def expand_and_flatten(self, val, cond, zcond):
        """Parse and expand a variable definition, flattening any
        recursive expansions by hoisting

        @return a Multiverse list of (cond, val) pairs"""
        # Parse the variable's definition
        d = parser.Data.fromstring(val, None)
        e, t, o = parser.parsemakesyntax(d, 0, (), parser.iterdata)
        if t != None or o != None:
            # TODO: do something if part of the string is left over
            pass

        expanded = self.process_expansion(e)
        # print("expanded", expanded)
        if isinstance(expanded, str):
            return Multiverse([CondDef(cond, zcond, expanded)])
        else: # must be a multiverse
            return expanded

    def join_values(self, value_list, delim=""):
        """Joins a list of make variable values that may be None, which
        means the variable is undefined.  When joined with defined values,
        None is the empty string.  When all values are None, the resulting
        value is also None.

        @param value_list a list of strings or Nones
        @returns A string or a None"""
        return delim.join(v for v in value_list if v)
    
    def append_values(self, *args):
        return self.join_values(args, " ")

    def bdd_to_dnf(self, condition):
        """Converts the expression to conjunctive normal form

        @returns a list of terms, represented as lists of factors"""

        solutions = kmaxtools.datastructures.bdd_solutions(condition)
        expression = []
        for solution_term in solutions:
            term = []
            for factor in solution_term:
                if solution_term[factor]:
                    term.append(self.reverse_bvars[factor])
                else:
                    term.append("!%s" % (self.reverse_bvars[factor]))
            expression.append(term)
        return expression

    def bdd_to_str(self, condition):
        """Converts the expression to a string"""

        expression = ""
        term_delim = ""
        for sublist in self.bdd_to_dnf(condition):
            term = ""
            factor_delim = ""
            for factor in sublist:
                term += factor_delim + factor
                factor_delim = " && "
            if len(term) > 0:
                expression += term_delim + term
                term_delim = " || "
        if len(expression) == 0:
            expression="1"
        return expression

    def add_var(self, name, presence_cond, presence_zcond, token, value):
        """Given a presence cond, the variable's name, its assignment
        operator, and the definition, update the variable's list of
        configurations to reflect the dry-runs needed to cover all its
        definition."""

        # print("add_var", name, z3.simplify(presence_zcond), token, value)

        assert value is not None, value

        update_vars = lambda name: \
                    map(lambda (old_value, old_cond, old_zcond, old_flavor): 
                            VarEntry(old_value, 
                                    conj(old_cond, neg(presence_cond)),
                                    zconj(old_zcond, z3.Not(presence_zcond)),
                                    old_flavor), 
                        self.variables[name])

        if token == "=":
            # Recursively-expanded variable defs are expanded at use-time

            equivs = self.get_var_equiv_set(name)
            for equiv_name in equivs:
                # Update all existing definitions' presence conds
                if equiv_name in self.variables:
                    self.variables[equiv_name] = update_vars(equiv_name)
                else:
                    self.variables[equiv_name] = []

                # Add complete definition to variable (needed to find variable
                # expansions at use-time)
                self.variables[equiv_name].append(
                    VarEntry(value, presence_cond, presence_zcond, VarEntry.RECURSIVE))

        elif token == ":=":
            # Simply-expanded self.variables are expanded at define-time

            equivs = self.get_var_equiv_set(name)
            for equiv_name in equivs:
                # Update all existing definitions' presence conds
                # AFTER getting the new definition, since the new
                # definition might refer to itself as in
                # linux-3.0.0/crypto/Makefile

                if equiv_name in self.variables:
                    old_variables = update_vars(equiv_name)
                else:
                    old_variables= [VarEntry("", 
                        neg(presence_cond), 
                        z3.Not(presence_zcond), 
                        VarEntry.RECURSIVE)]
                    # old_variables = []

                # Expand and flatten self.variables in the definition and add the
                # resulting definitions.
                new_definitions = self.expand_and_flatten(value, presence_cond, presence_zcond)
                # print equiv_name
                # print new_definitions
                new_variables = []
                for new_cond, new_zcond, new_value in new_definitions:
                    new_variables.append(VarEntry(new_value, 
                                                    new_cond,
                                                    new_zcond,
                                                    VarEntry.SIMPLE))

                self.variables[equiv_name] = old_variables + new_variables
                # TODO: check for computed variable names, compute them, and
                # collect any configurations resulting from those self.variables

        elif token == "+=":
            # optimize append for certain variables this
            # optimization creates a new variable entry for
            # append instead of computing the cartesian
            # product of appended variable definitions
            simply = self.F
            zsimply = ZSolver.F
            recursively = self.F
            zrecursively = ZSolver.F

            new_var_name = self.get_var_equiv(name)

            # find conds for recursively- and simply-expanded variables
            if name in self.variables:
                for entry in self.variables[name]:
                    _, old_cond, old_zcond, old_flavor = entry
                    # TODO: optimization, memoize these
                    if old_flavor == VarEntry.SIMPLE:
                        simply = disj(simply, old_cond)
                        zsimply = zdisj(zsimply, old_zcond)
                    else:
                        assert old_flavor == VarEntry.RECURSIVE
                        recursively = disj(recursively, old_cond)
                        zrecursively = zdisj(zrecursively, old_zcond)

                # print("+=", name, z3.simplify(zsimply), z3.simplify(zrecursively))

                if not isfalse(recursively, zrecursively):
                    if new_var_name not in self.variables:
                        self.variables[new_var_name] = []
                    self.variables[new_var_name].append(
                        VarEntry(value,
                                presence_cond,
                                presence_zcond,
                                VarEntry.RECURSIVE))

                if not isfalse(simply, zsimply):
                    new_definitions = self.expand_and_flatten(value, presence_cond,presence_zcond)
                    # print("simply", new_definitions)
                    new_variables = []
                    for new_cond, new_zcond, new_value in new_definitions:
                        and_cond = conj(new_cond, presence_cond)
                        and_zcond = zconj(new_zcond, presence_zcond)
                        if not isfalse(and_cond, and_zcond):
                            new_variables.append(VarEntry(
                                new_value, and_cond, and_zcond, VarEntry.SIMPLE))

                    self.variables[new_var_name] = new_variables
                    # print("simply_done", new_variables)


            else:
                self.variables[new_var_name] = [VarEntry(
                    value, presence_cond, presence_zcond, VarEntry.RECURSIVE)]       
                
                    
        elif token == "?=":
            # TODO: if ?= is used on an undefined variable, it's
            # initialized as a recursively-expanded variable (=)

            pass
        
        else:
            mlog.error("Unknown setvariable token: {}".format(token))

        # Trim definitions with a presence cond of FALSE                    
        for equiv in self.get_var_equiv_set(name):
            if equiv in self.variables:
                self.variables[equiv] = \
                    [v for v in self.variables[equiv] if not isfalse(v.cond, v.zcond)]
                

    def process_setvariable(self, setvar, cond, zcond):
        """Find a satisfying set of configurations for variable."""
        # assert isinstance(cond, pycudd.DdNode), cond
        assert isinstance(setvar, parserdata.SetVariable), setvar
        assert z3.is_expr(zcond), zcond

        # obj-y = 'fork.o'
        name = self.process_expansion(setvar.vnameexp)
        token = setvar.token
        value = setvar.value

        # def f(s, cond, zcond):
        #     if not (s.endswith("-y") or s.endswith("-m")):
        #         return

        #     vs = [v for v in value.split()
        #           if v.endswith(".o") or v.endswith("/")]

        #     for v in vs:
        #         if v not in self.token_pc:
        #             self.token_pc[v] = (self.T, ZSolver.T)

        #         # update nested_cond
        #         tc, tzc = self.token_pc[v]
        #         self.token_pc[v] = (conj(tc, cond), zconj(tzc, zcond))
                
        #         if s not in set(["obj-y", "obj-m", "lib-y", "lib-m"]):
        #             continue
                
        #         if v.endswith(".o"):
        #             if v in self.unit_pc:
        #                 uc, uzc = self.unit_pc[v]
        #                 tc, tzc = self.token_pc[v]
        #                 self.unit_pc[v] = (disj(uc, tc), zdisj(uzc, tzc))
        #             else:
        #                 self.unit_pc[v] = self.token_pc[v]
        #         elif v.endswith("/"):
        #             assert v not in self.subdir_pc, (v, self.subdir_pc)
        #             self.subdir_pc[v] = self.token_pc[v]
                

        if isinstance(name, str):
            # f(name, cond, zcond)  # remove because this method for presence conditions is obsolete
            self.add_var(name, cond, zcond, token, value)
        else:
            for local_cond, local_zcond, expanded_name in name:
                nested_cond = conj(local_cond, cond)
                nested_zcond = zconj(local_zcond, zcond)
                # f(expanded_name, nested_cond, nested_zcond)  # remove because this method for presence conditions is obsolete
                self.add_var(expanded_name, nested_cond, nested_zcond, token, value)

    def process_rule(self, rule, cond, zcond):
        # mlog.warn("just pass on rule {} {} {}".format(rule, self.bdd_to_str(cond), zcond))
        mlog.warn("just pass on rule {}".format(rule))
        pass

    def process_include(self, s, cond, zcond):
        expanded_include = self.mk_Multiverse(self.process_expansion(s.exp))
        for include_cond, include_zcond, include_files in expanded_include:
            if include_files != None:
                for include_file in include_files.split():
                    obj = os.path.dirname(include_file)
                    if os.path.exists(include_file):
                        include_makefile = open(include_file, "rU")
                        s = include_makefile.read()
                        include_makefile.close()
                        include_stmts = parser.parsestring(s, include_makefile.name)
                        self.process_stmts(include_stmts, include_cond, include_zcond)

    def split_defs(self, var):
        """get every whitespace-delimited token in all definitions of the
        given var name, expanding any var invocations first"""

        if var not in self.variables:
            return []

        values = []
        for (value, cond, zcond, _) in self.variables[var]:
            assert value is not None, value
            # Expand any vars used in definitions

            expanded_values = self.mk_Multiverse(
                self.expand_and_flatten(value, cond, zcond))

            for expanded_cond, expanded_zcond, expanded_value in expanded_values:
                if expanded_value is None:
                    continue
                
                split_expanded_values = expanded_value.split()
                values.extend(split_expanded_values)

                # composite_unit = "-".join(var.split('-')[:-1]) + ".o"
                # if composite_unit not in self.token_pc:
                #     continue
                
                # for v in split_expanded_values:
                #     tc, tzc = self.token_pc[composite_unit]
                #     composite_cond = conj(expanded_cond, tc)
                #     composite_zcond = zconj(expanded_zcond, tzc)

                #     if not v in self.token_pc.keys():
                #         self.token_pc[v] = self.T, ZSolver.T

                #     # update nested_cond
                #     tc, tzc = self.token_pc[v]
                #     self.token_pc[v] = (conj(tc, composite_cond),
                #                         zconj(tzc, composite_zcond))

        return values
                

class Run:    

    def run(self, makefiledirs):
        assert isinstance(makefiledirs, (set, list)) \
            and makefiledirs, makefiledirs
        assert all(os.path.isfile(f) or os.path.isdir(f)
                   for f in makefiledirs), makefiledirs

        self.results = Results()

        subdirs = set(makefiledirs)
        while subdirs:
            makefile = subdirs.pop()
            mlog.info("processing makefile: {}".format(makefile))            
            subdirs_ = self.extract(makefile)
            if kmaxtools.settings.do_recursive:
                subdirs = subdirs.union(subdirs_)

    def extract(self, path):
        makefile = self.get_makefile(path)

        path = os.path.dirname(makefile)
        makefile = open(makefile, "rU")

        s = makefile.read()
        makefile.close()

        kbuild = Kbuild()
        kbuild.add_definitions(kmaxtools.settings.defines)
        stmts = parser.parsestring(s, makefile.name)

        kbuild.process_stmts(stmts, kbuild.T, ZSolver.T)
        # SPECIAL-obj-simple uses a simply-expanded variable to expand obj-y in case obj-y is recursively-expanded, which means the variables haven't been expanded in obj-y yet, e.g., ptrace_$(BITS)
        kbuild.process_stmts(parser.parsestring("SPECIAL-obj-simple := $(obj-y) $(obj-m)", makefile.name), kbuild.T, ZSolver.T)
        kbuild.process_stmts(parser.parsestring("SPECIAL-core-simple := $(core-y) $(core-m) $(drivers-y) $(drivers-m) $(net-y) $(net-m) $(libs-y) $(libs-m) $(head-y) $(head-m)", makefile.name), kbuild.T, ZSolver.T)
        
        subdirs = self.results.subdirs
        compilation_units = self.results.compilation_units
        composites = self.results.composites
        library_units = self.results.library_units
        hostprog_composites = self.results.hostprog_composites
        hostprog_units = self.results.hostprog_units
        unconfigurable_units = self.results.unconfigurable_units
        # unit_pcs = self.results.unit_pcs
        # subdir_pcs = self.results.subdir_pcs
        clean_files = self.results.clean_files

        pending_vars = set(
            ["obj-y", "obj-m",
             "core-y", "core-m", "drivers-y", "drivers-m",
             "net-y", "net-m", "libs-y", "libs-m", "head-y",
             "head-m", "SPECIAL-obj-simple", "SPECIAL-core-simple"])
             
        self.collect_units(kbuild,
                           path,
                           pending_vars,
                           compilation_units,
                           subdirs,
                           composites)

        for v in set([ "subdir-y", "subdir-m" ]):
            for u in kbuild.split_defs(v):
                subdirs.add(os.path.join(path, u))

        self.collect_units(kbuild,
                           path,
                           set(["lib-y", "lib-m" ]),
                           library_units,
                           subdirs,
                           composites)

        pending_hostprog_composites = set()
        for v in set([ "hostprogs-y", "hostprogs-m", "host-progs", "always" ]):
            for u in kbuild.split_defs(v):
                composite_name = u + "-objs"
                unit_name = os.path.join(path, u)
                if composite_name in kbuild.variables:
                    pending_hostprog_composites.add(composite_name)
                    hostprog_composites.add(unit_name)
                else:
                    hostprog_units.add(unit_name)

        if pending_hostprog_composites:
            raise NotImplementedError
            self.collect_units(kbuild,
                               path,
                               pending_hostprog_composites,
                               hostprog_units,
                               None,
                               hostprog_composites)

        # todo: collect non-kbuild object files to replicate prior analyses
        # for v in set([ "targets", "extra-y" ]):
        #     for u in kbuild.split_defs(v):
        #         unit_name = os.path.join(path, u)
        #         if unit_name.endswith(".o"):
        #             raise NotImplementedError
        #             # extra_targets.add(unit_name)

        # for u in kbuild.split_defs("targets"):
        #     if u.endswith(".c"):
        #         raise NotImplementedError
        #         #c_file_targets.add(os.path.join(obj, u))

        # for u in kbuild.split_defs("clean-files"):
        #     clean_files.add(os.path.join(path, u))
            

        # look for variables starting with obj-, lib-, hostprogs-,
        # compositename-, or hostprog-compositename-.  doesn't account for
        # composites in the unconfigurable variables
        unconfigurable_prefixes = set([ "obj-$", "lib-$", "hostprogs-$" ])
        for cset in (composites, hostprog_composites):
            for c in cset:
                c = os.path.basename(c)
                if c.endswith(".o"):
                    c = c[:-2]
                unconfigurable_prefixes.add(c + "-$")
        unconfigurable_variables = set([])
        for x in kbuild.variables:
            for p in unconfigurable_prefixes:
                if x.startswith(p) and \
                        not x.endswith("-") and \
                        not x.endswith("-y") and \
                        not x.endswith("-m") and \
                        not x.endswith("-objs") and \
                        x != "host-progs":
                    unconfigurable_variables.add(x)
                elif x.startswith(p[:-1]) and x.endswith("-"):
                    # also look in variables resulting from expansion of
                    # undefined config var
                    unconfigurable_variables.add(x)
        self.collect_units(kbuild,
                           path,
                           unconfigurable_variables,
                           unconfigurable_units,
                           unconfigurable_units,
                           unconfigurable_units)

        # subtract out compilation units that are configurable
        unconfigurable_units.difference_update(compilation_units)
        unconfigurable_units.difference_update(library_units)
        unconfigurable_units.difference_update(composites)
        unconfigurable_units.difference_update(subdirs)

        # look for variable expansions or function calls in
        # compilation_units, subdirs, and variable names
        self.check_unexpanded_vars(compilation_units, "compilation unit")
        self.check_unexpanded_vars(subdirs, "subdirectory")
        self.check_unexpanded_vars(kbuild.variables.keys(), "variable name")

        if kmaxtools.settings.do_table:
            mlog.info(kbuild.getSymbTable(printCond=kbuild.bdd_to_str))

        presence_conditions = {}
        kbuild.get_presence_conditions([ "obj-y", "obj-m", "lib-y", "lib-m", "SPECIAL-obj-simple" ], presence_conditions, kbuild.T, ZSolver.T)
        for token in presence_conditions:
            # resolve any uses of ../ or ./
            filename = os.path.join(path, token)
            if filename not in self.results.presence_conditions.keys():
                self.results.presence_conditions[filename] = presence_conditions[token]
            else:
                self.results.presence_conditions[filename] = zdisj(self.results.presence_conditions[filename], presence_conditions[token])

        presence_conditions = {}
        kbuild.get_presence_conditions([ "core-y", "core-m",
                                         "drivers-y", "drivers-m", "net-y", "net-m", "libs-y",
                                         "libs-m", "head-y", "head-m", "SPECIAL-core-simple"], presence_conditions,
                                       kbuild.T, ZSolver.T)
        for token in presence_conditions:
            filename = token
            if filename not in self.results.presence_conditions.keys():
                self.results.presence_conditions[filename] = presence_conditions[token]
            else:
                self.results.presence_conditions[filename] = zdisj(self.results.presence_conditions[filename], presence_conditions[token])

        # removed because this method for getting presence conditions is obsolete
        # def _f(d, s):
        #     for name, (cond, zcond) in d.iteritems():
        #         s.add((os.path.normpath(os.path.join(path, name)), kbuild.bdd_to_str(cond), zcond))

        # _f(kbuild.unit_pc, unit_pcs)
        # _f(kbuild.subdir_pc, subdir_pcs)

        #clean up
        bdd_destroy()
        
        return subdirs

    @classmethod
    def get_makefile(cls, path):
        #use Kbuild file if found, otherwise try Makefile
        if os.path.isdir(path):
            makefile = os.path.join(path, "Kbuild")
            if not os.path.isfile(makefile):
                makefile = os.path.join(path, "Makefile")

                if not os.path.isfile(makefile):
                    mlog.error("{} not found".format(makefile))
                    exit(1)
        else:
            assert os.path.isfile(path), path
            makefile = path                

        return makefile

    @classmethod
    def collect_units(cls,
                      kbuild,            # current kbuild instance
                      path,               # current path
                      pending_vars, # variables to extract from, gets emptied
                      compilation_units, # adds units to this set
                      subdirs,    # adds subdir to this set
                      composites):  # save pcs from these variables

        """fixed-point algorithm that adds composites and stops when no
        more variables to look at are available"""

        assert isinstance(kbuild, Kbuild), kbuild
        assert path is None or path is "" or os.path.isdir(path), path
        assert isinstance(compilation_units, set), compilation_units
        assert isinstance(subdirs, set), subdirs
        assert isinstance(composites, set), composites

        equiv_sets = [kbuild.get_var_equiv_set(v) for v in pending_vars]
        equiv_vars = set.union(*equiv_sets) if equiv_sets else set()
        pending_vars |= equiv_vars

        processed_vars = set()        
        while pending_vars:
            pending_var = pending_vars.pop()
            processed_vars.add(pending_var)

            values = kbuild.split_defs(pending_var)
            for elem in values:
                cls.collect_defs(elem, kbuild, path,
                                 compilation_units,
                                 subdirs, composites, 
                                 processed_vars, pending_vars)
                
    @classmethod
    def collect_defs(cls, elem,
                     kbuild,
                     path,
                     compilation_units,
                     subdirs,
                     composites,
                     processed_vars,
                     pending_vars):

        unit_name = os.path.normpath(os.path.join(path, elem))
        if elem.endswith(".o") and unit_name not in compilation_units:
            # Expand composites
            if (elem[:-2] + "-objs") in kbuild.variables or \
                (elem[:-2] + "-y") in kbuild.variables:
                # scripts/Makefile.build use the -objs and -y
                # suffix to define composites $($(subst
                # $(obj)/,,$(@:.o=-objs))) $($(subst
                # $(obj)/,,$(@:.o=-y)))), $^)
                composite_variable1 = elem[:-2] + "-objs"
                composite_variable2 = elem[:-2] + "-y"

                if composite_variable1 not in processed_vars and \
                        composite_variable2 not in processed_vars:
                    composites.add(unit_name)
                    pending_vars.update(kbuild.get_var_equiv_set(composite_variable1))
                    pending_vars.update(kbuild.get_var_equiv_set(composite_variable2))
                    # if (elem not in kbuild.token_pc):
                    #     raise NotImplementedError
                    #     kbuild.token_pc[elem] = (kbuild.T, ZSolver.T)
                    # kbuild.composite_pc[elem] = kbuild.token_pc[elem]

                if os.path.isfile(unit_name[:-2] + ".c") or os.path.isfile(unit_name[:-2] + ".S"): 
                    compilation_units.add(unit_name)
                    # if (elem not in kbuild.token_pc): 
                    #     kbuild.token_pc[elem] = (kbuild.T, ZSolver.T)
                    # kbuild.unit_pc[elem] = kbuild.token_pc[elem]
            else:
                compilation_units.add(unit_name)
                # if (elem not in kbuild.token_pc):
                #     kbuild.token_pc[elem] = (kbuild.T, ZSolver.T)
                # kbuild.unit_pc[elem] = kbuild.token_pc[elem]

        elif elem.endswith("/"):
            # scripts/Makefile.lib takes anything that
            # ends in a forward slash as a subdir
            # $(patsubst %/,%,$(filter %/, $(obj-y)))
            new_dir = elem
            if not new_dir.startswith('/'):
                new_dir = os.path.join(path, new_dir)

            if os.path.isdir(new_dir):
                subdirs.add(new_dir)

            # if elem not in kbuild.token_pc:
            #     kbuild.token_pc[elem] = kbuild.T
            # kbuild.subdir_pc[elem] = kbuild.token_pc[elem]
    
    @classmethod
    def check_unexpanded_vars(cls, l, desc):
        for x in l:
            if contains_unexpanded(x):
                mlog.warn("A {} contains an unexpanded variable or call {}".format(desc, x))



import traceback
import sys
import ast
import z3
import regex

# Documentation of ast and the visitors: https://greentreesnakes.readthedocs.io/en/latest/nodes.html

t_simplify = z3.Tactic('ctx-solver-simplify')
t_tseitin = z3.Tactic('tseitin-cnf-core')
identifier_pattern = regex.compile("^[A-Za-z0-9_]+$") # config var names can start with numbers, see kconfig/lexer.l
int_pattern = regex.compile("^[0-9]+$")
hex_pattern = regex.compile("^0x[0-9ABCDEFabcdef]+$")

def glean_unknown_symbol(sym):
  sym = str(sym)
  # sys.stderr.write("trying to glean unknown symbol: \"%s\"\n" % (sym))
  if int_pattern.match(sym):
    num = int(sym)
    if num == 0:
      return z3.BoolVal(False)
    else:
      return z3.BoolVal(True)
  elif hex_pattern.match(sym):
    return z3.Bool("\"%s\"" % (sym))
  # elif identifier_pattern.match(sym):
  #   return z3.Bool("CONFIG_%s" % (sym))
  else:
    return z3.StringVal(sym)

# parse tree processing
class Converter(ast.NodeVisitor):
  def __init__(self):
    ast.NodeVisitor.__init__(self)
    self.z3 = None

  def result(self):
    return self.z3

  def visit_Expr(self, node):
    self.generic_visit(node)
    node.z3 = node.value.z3
    self.z3 = node.z3

  def visit_BoolOp(self, node):
    self.generic_visit(node)
    operands = []
    for value in node.values:
      operands.append(value.z3)
    # string values should not normally appear in boolean expressions,
    # but this can happen when there are undefined configuration
    # options used in expressions.  instead of throwing an error,
    # kconfig treats the usage of undefined configuration options
    # (identifiers) as strings.  such strings are assumed to be false,
    # since they aren't options that can ever be true.  (see visit_UnaryOp)
    operands = [(operand if z3.is_bool(operand) else z3.BoolVal(False)) for operand in operands]
    if node.op.name == "and":
      node.z3 = z3.And(operands)
    elif node.op.name == "or":
      node.z3 = z3.Or(operands)
    else:
      assert(False)

  def visit_And(self, node):
    node.name = "and"

  def visit_Or(self, node):
    node.name = "or"

  def visit_Not(self, node):
    node.name = "not"

  def visit_UnaryOp(self, node):
    self.generic_visit(node)
    operand = node.operand.z3
    # print("lalalal")
    # print(operand)
    if node.op.name == "not":
      if z3.is_bool(operand):
        node.z3 = z3.Not(operand)
      else:
        # the expression could be a stringval in the case of the usage
        # of an undefined configuration option, which kconfig allows.
        # we treat such variables are always off, so negating that
        # variable makes it always true.  we can't assume all strings
        # are undefined variables, because string constants are also
        # permitted in kconfig.  such string constants have a false
        # value anyway.  (see visit_BoolOp)
        node.z3 = z3.BoolVal(True)
    else:
      assert(False)

  def visit_Constant(self, node):
    # sys.stderr.write("const %s %s\n" % (str(node), str(type(node.value))))
    value = node.value
    result = glean_unknown_symbol(value)
    # sys.stderr.write("result const: %s\n" % (result))
    if result is not None:
      node.z3 = result
    else:
      print(traceback.format_exc())
      sys.stderr.write("error: cannot process value \"%s\"\n" % (value))
      exit(1)
      node.z3 = None

  def visit_Num(self, node):  # deprecated since 3.8, replaced by constant
    num = node.n
    if num == 0:
      node.z3 = z3.BoolVal(False)
    else:
      node.z3 = z3.BoolVal(True)

  def visit_Str(self, node):  # deprecated since 3.8, replaced by constant
    value = node.s
    result = glean_unknown_symbol(value)
    if result is not None:
      node.z3 = result
    else:
      print(traceback.format_exc())
      sys.stderr.write("error: cannot process value \"%s\"\n" % (value))
      exit(1)
      node.z3 = None

  def visit_Name(self, node):
    # sys.stderr.write("name %s\n" % (str(node)))
    node.z3 = z3.Bool(node.id)
    # sys.stderr.write("name z3 %s\n" % (str(result)))
    # sys.stderr.write("name is_bool %s\n" % (str(z3.is_bool(result))))
    
  def visit_Compare(self, node):
    self.generic_visit(node)
    if len(node.ops) > 1:
      # kconfig constraints aren't expected to have multiple operators in a comparison expression
      assert(False)
    left = node.left.z3
    op = node.ops[0].name
    right = node.comparators[0].z3

    # model non-Boolean operators by creating a predicate variable.
    # TODO: model relational expressions and expressions between variables and strings in Z3
    predicate = str(ast.dump(node))
    node.z3 = z3.Bool("PREDICATE_%s" % (predicate))

    # represent (in)equality between booleans and strings using z3 expressions
    if z3.is_string(left) and z3.is_string(right) or z3.is_bool(left) and z3.is_bool(right):
      if op == "eq":
        node.z3 = z3.simplify(z3.Not(z3.Distinct(left, right)))
      elif op == "ne":
        node.z3 = z3.simplify(z3.Distinct(left, right))
      # elif op == "lt":
      #   node.z3 = z3.Z3_mk_lt(z3.get_ctx(None), left.ast, right.ast)
      # elif op == "le":
      #   node.z3 = z3.simplify(z3.ULE(left, right))
      # elif op == "gt":
      #   node.z3 = z3.simplify(z3.UGT(left, right))
      # elif op == "ge":
      #   node.z3 = z3.simplify(z3.UGE(left, right))
      # else:
      #  assert(False)

  def visit_Eq(self, node):
    node.name = "eq"

  def visit_NotEq(self, node):
    node.name = "ne"

  def visit_Lt(self, node):
    node.name = "lt"

  def visit_LtE(self, node):
    node.name = "lte"

  def visit_Gt(self, node):
    node.name = "gt"

  def visit_GtE(self, node):
    node.name = "gte"

  # def default(self, node):
  #   # sys.stderr.write("trying to process default: \"%s\"\n" % (node))
  #   result = glean_unknown_symbol(node)
  #   # sys.stderr.write("result: %s\n" % (result))
  #   if result is not None:
  #     return result
  #   else:
  #     predicate = str(node)
  #     return z3.Bool("PREDICATE_%s" % (predicate))


# parse tree processing
class IdentifierCollector(ast.NodeVisitor):
  def __init__(self):
    self.identifiers = []

  def result(self):
    return self.identifiers

  def visit_Name(self, node):
    self.identifiers.append(node.id)

def get_identifiers(expr):
  try:
    tree = ast.parse(expr)
  except:
    sys.stderr.write("error: could not parse %s\n" % (line))
    sys.stderr.write(traceback.format_exc())
    sys.stderr.write("\n")
    return []
  visitor = IdentifierCollector()
  visitor.visit(tree)
  return visitor.result()

def convert_to_z3(expr):
  # print(expr)
  try:
    tree = ast.parse(expr)
  except RuntimeError as e:
    sys.stderr.write("exception: %s\n" % (e))
    sys.stderr.write("could not convert expression to cnf\n%s\n" % (expr))
    exit(1)
    return []
  except Error as e:
    # handle broken expressions
    sys.stderr.write("exception: %s\n" % (e))
    sys.stderr.write("warning: couldn't parse %s\n" % (expr))
    exit(1)
    return []
  # sys.stderr.write("actual_expr %s\n" % (str(actual_expr)))
  visitor = Converter()
  # print(ast.dump(tree))
  visitor.visit(tree)
  z3_clause = visitor.result()
  # sys.stderr.write("z3_clause %s\n" % (str(z3_clause)))
  return z3.simplify(z3_clause)

if __name__ == '__main__':
  # tests
  print(convert_to_z3('((((not ((((CONFIG_GCOV_KERNEL and CONFIG_CC_IS_GCC and 1)) and (((1 and CONFIG_GCC_VERSION < "40700")) or ((1)))))) or (CONFIG_GCOV_FORMAT_3_4 or CONFIG_GCOV_FORMAT_4_7))) and (((not (CONFIG_GCOV_FORMAT_3_4 or CONFIG_GCOV_FORMAT_4_7)) or ((CONFIG_GCOV_KERNEL and CONFIG_CC_IS_GCC and 1)))))'))
  # print(convert_to_z3('((((not ((((1)) and (((1)) or ((1 and CONFIG_GCC_PLUGINS)) or ((1 and CONFIG_GCC_PLUGINS and ( not CONFIG_KASAN or CONFIG_KASAN_STACK!="1"))) or ((1 and CONFIG_GCC_PLUGINS and ( not CONFIG_KASAN or CONFIG_KASAN_STACK!="1"))) or ((1 and CONFIG_CC_HAS_AUTO_VAR_INIT)))))) or (CONFIG_INIT_STACK_NONE or CONFIG_GCC_PLUGIN_STRUCTLEAK_USER or CONFIG_GCC_PLUGIN_STRUCTLEAK_BYREF or CONFIG_GCC_PLUGIN_STRUCTLEAK_BYREF_ALL or CONFIG_INIT_STACK_ALL))) and (((not (CONFIG_INIT_STACK_NONE or CONFIG_GCC_PLUGIN_STRUCTLEAK_USER or CONFIG_GCC_PLUGIN_STRUCTLEAK_BYREF or CONFIG_GCC_PLUGIN_STRUCTLEAK_BYREF_ALL or CONFIG_INIT_STACK_ALL)) or ((1)))))'))
  # print(convert_to_z3('((((not ((((1)) and (((1 and CONFIG_X86_64)) or ((1)) or ((1 and CONFIG_EXPERT and  not CONFIG_STACKDEPOT)))))) or (CONFIG_UNWINDER_ORC or CONFIG_UNWINDER_FRAME_POINTER or CONFIG_UNWINDER_GUESS))) and (((not (CONFIG_UNWINDER_ORC or CONFIG_UNWINDER_FRAME_POINTER or CONFIG_UNWINDER_GUESS)) or ((1)))))'))
  # print(get_identifiers('((((not ((((1)) and (((1 and CONFIG_X86_64)) or ((1)) or ((1 and CONFIG_EXPERT and  not CONFIG_STACKDEPOT)))))) or (CONFIG_UNWINDER_ORC or CONFIG_UNWINDER_FRAME_POINTER or CONFIG_UNWINDER_GUESS))) and (((not (CONFIG_UNWINDER_ORC or CONFIG_UNWINDER_FRAME_POINTER or CONFIG_UNWINDER_GUESS)) or ((1)))))'))



  



import os
from setuptools import setup, Extension

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

about = {}
exec(read(os.path.join("kmaxtools", "about.py")), about)

kextractor_next_20200430 = Extension('kextractor_next_20200430', [ 'kextractors/kextractor-next-20200430/kextractor_extension.c', 'kextractors/kextractor-next-20200430/kextractor.c', 'kextractors/kextractor-next-20200430/confdata.c', 'kextractors/kextractor-next-20200430/expr.c', 'kextractors/kextractor-next-20200430/preprocess.c', 'kextractors/kextractor-next-20200430/lexer.lex.c', 'kextractors/kextractor-next-20200430/parser.tab.c', 'kextractors/kextractor-next-20200430/symbol.c', 'kextractors/kextractor-next-20200430/util.c'], include_dirs=['kextractors/kextractor-next-20200430/'])

kextractor_3_19 = Extension('kextractor_3_19', [ 'kextractors/kextractor-3.19/kextractor_extension.c', 'kextractors/kextractor-3.19/kextractor.c', 'kextractors/kextractor-3.19/bconf.tab.c', 'kextractors/kextractor-3.19/zconf.tab.c'], include_dirs=['kextractors/kextractor-3.19/'])

kextractor_4_12_8 = Extension('kextractor_4_12_8', [ 'kextractors/kextractor-4.12.8/kextractor_extension.c', 'kextractors/kextractor-4.12.8/kextractor.c', 'kextractors/kextractor-4.12.8/bconf.tab.c', 'kextractors/kextractor-4.12.8/zconf.tab.c'], include_dirs=['kextractors/kextractor-4.12.8/'])

setup(
    name = about['__title__'],
    version = about['__version__'],
    author = "Paul Gazzillo",
    author_email = "paul@pgazz.com",
    description = ("Tools for working with symbolic  constraints from Kbuild Makefile."),
    long_description_content_type = 'text/markdown',
    long_description = read('README.md'),
    license = "GPLv2+",
    keywords = "makefile kconfig kbuild configurations kmax kclause klocalizer",
    url = "https://github.com/paulgazz/kmax",
    packages=['kmaxtools', 'pymake'],
    ext_modules = [ kextractor_next_20200430, kextractor_3_19, kextractor_4_12_8 ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
    ],
    scripts=['kmaxtools/kmax', 'kmaxtools/kmaxall', 'kmaxtools/kclause', 'kmaxtools/klocalizer', 'kmaxtools/kextract'],
    install_requires=[
        'enum34',
        'regex',
        'z3-solver',
        'dd',
        'networkx==2.2', # for dd to work on python2
    ],
    use_2to3=True,
)
